#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.
import os
import getpass
import subprocess
import pytest

from sos import execute_workflow
from sos._version import __version__
from sos.utils import env, load_config_files
from sos.eval import get_config

# if the test is imported under sos/test, test interacive executor

test_cfg = '''
    cut: 0.5
    cut1:
    - 0.5
    - 2
    - 3
    cut2: a3
    cut3:
    - a
    - b
    - c
    cut4:
    A: 123
    me: '{user_name}@my'
'''


def test_command_line():
    '''Test command line arguments'''
    assert subprocess.call(
        'sos config -h',
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        shell=True) == 0
    assert subprocess.call(
        'sos config -g --get',
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        shell=True) == 0
    assert subprocess.call(
        'sos config',
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        shell=True) == 0
    assert subprocess.call(
        'sos config --get',
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        shell=True) == 0
    assert subprocess.call(
        'sos config -g --set a 5',
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        shell=True) == 0
    assert subprocess.call(
        'sos config --get a',
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        shell=True) == 0
    assert subprocess.call(
        'sos config -g --unset a',
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        shell=True) == 0


def test_config_set(config_factory):
    '''Test interpolation of config'''
    myconfig = config_factory(test_cfg)

    assert subprocess.call(
        f'sos config --set cut 0.5 -c {myconfig}',
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        shell=True) == 0
    load_config_files(myconfig)
    assert env.sos_dict['CONFIG']['cut'] == 0.5
    #
    assert subprocess.call(
        f'sos config --set cut1 0.5 2 3 -c {myconfig}',
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        shell=True) == 0
    load_config_files(myconfig)
    assert env.sos_dict['CONFIG']['cut1'] == [0.5, 2, 3]
    #
    assert subprocess.call(
        f'sos config --set cut2 a3 -c {myconfig}',
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        shell=True) == 0
    load_config_files(myconfig)
    assert env.sos_dict['CONFIG']['cut2'] == 'a3'
    #
    assert subprocess.call(
        f'sos config --set cut3 a b c -c {myconfig}',
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        shell=True) == 0
    load_config_files(myconfig)
    assert env.sos_dict['CONFIG']['cut3'] == ['a', 'b', 'c']
    #
    assert subprocess.call(
        f'''sos config --set cut4 "{{'A': 123}}" -c {myconfig}''',
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        shell=True) == 0
    load_config_files(myconfig)
    assert env.sos_dict['CONFIG']['cut4'] == {'A': 123}


def test_interpolate(config_factory):
    '''Test interpolation of config'''
    myconfig = config_factory(test_cfg)
    assert subprocess.call(
        f'''sos config --set me '{{user_name}}@my' -c {myconfig}''',
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        shell=True) == 0
    load_config_files(myconfig, default_config_files=False)
    assert get_config('me') == f'{getpass.getuser().lower()}@my'


def test_global_vars(config_factory):
    '''Test SoS defined variables'''
    execute_workflow("[0]", options={'mode': 'dryrun'})

    assert env.sos_dict['SOS_VERSION'] == __version__
    assert isinstance(env.sos_dict['CONFIG'], dict)

    cfg = config_factory({'my_config': 5})

    execute_workflow("[0]", options={'config_file': cfg})
    assert env.sos_dict['CONFIG']['my_config'] == 5


def test_get_config(config_factory):
    myconfig = config_factory({
        'val': 5,
        'A': {
            'B.C': '33',
            'B.C1': {
                'D': '34'
            },
            'D': '45'
        },
        'E': {
            'F': {
                'val': 6,
                'val1': 10,
                'G': '{val + val1}'
            },
            'H': '{val}'
        },
        'O': 'A{nonexisting}',
        'X': '{os.environ.get("HOME", "no_home")}'
    })
    load_config_files(myconfig)
    assert get_config('A', 'D') == '45'
    assert get_config('A.D') == '45'
    assert get_config(['A', 'D']) == '45'
    assert get_config(['A', 'D']) == '45'
    assert get_config('A.B.C') == '33'
    assert get_config('A.B.C1.D') == '34'
    assert get_config('A') == {'B.C': '33', 'B.C1': {'D': '34'}, 'D': '45'}
    assert get_config('E.F') == {'val': 6, 'val1': 10, 'G': '16'}
    assert get_config('E.F', val=7) == {'val': 6, 'val1': 10, 'G': '17'}
    assert get_config('E.F', val=7, allowed_keys=['G']) == {'G': '17'}
    assert get_config(
        'E.F', val=7, val1=20) == {
            'val': 6,
            'val1': 10,
            'G': '27'
        }
    assert get_config('E.F', {
        'val': 8,
        'val1': 30
    }) == {
        'val': 6,
        'val1': 10,
        'G': '38'
    }
    assert get_config('E.H', val=7) == '7'
    with pytest.raises(ValueError):
        get_config('O')
    assert get_config('O', nonexisting=7) == 'A7'
    assert get_config('X') == os.environ.get("HOME", "no_home")



#!/usr/bin/env python
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import glob
import os
import shutil
import socket
import sys
import time
import platform
import pytest

from sos import execute_workflow
from sos.parser import SoS_Script
from sos.targets import file_target
from sos.utils import env


def internet_on(host="8.8.8.8", port=80, timeout=3):
    """Test if internet is connected """
    try:
        socket.setdefaulttimeout(timeout)
        socket.create_connection(("www.google.com", 80))
        return True
    except Exception as e:
        print(e)
        return False


with_network = internet_on()


def test_acceptable_args():
    """test acceptable args of options"""
    with pytest.raises(Exception):
        execute_workflow(
            r"""
            run: unrecog=1
            echo 'a'
            """
        )


def test_get_output():
    """Test utility function get_output"""
    execute_workflow(
        r"""
        [0: shared='ret']
        ret = get_output('echo blah')
        """
    )
    # use strip because there would be \r\n under windows
    assert env.sos_dict["ret"].strip() == "blah"

    execute_workflow(
        r"""
        [0: shared='ret']
        ret = get_output('echo blah', show_command=True)
        """
    )
    assert [x.strip() for x in env.sos_dict["ret"].splitlines()] == [
        "$ echo blah",
        "blah",
    ]

    execute_workflow(
        r"""
        [0: shared='ret']
        ret = get_output('echo blah', show_command=True, prompt='% ')
        """
    )
    assert [x.strip() for x in env.sos_dict["ret"].splitlines()] == [
        "% echo blah",
        "blah",
    ]

    with pytest.raises(Exception):
        execute_workflow(
            r"""
            [0]
            get_output('catmouse')
        """
        )

    with pytest.raises(Exception):
        execute_workflow(
            r"""
            [0]
            ret = get_output('cat -h')
            """
        )


@pytest.mark.skipif(
    not shutil.which("bash") or sys.platform == "win32", reason="Needs bash."
)
def test_get_output_extra_kwargs():
    execute_workflow(
        r"""
        [0]
        ret = get_output('echo "ECHO" | wc -l', executable="/bin/bash")
        """
    )


def test_fail_if(temp_factory):
    """Test action fail if"""
    temp_factory("a.txt")

    # should fail in dryrun mode
    with pytest.raises(Exception):
        execute_workflow(
            r"""
            [0]
            input: 'a.txt'
            fail_if(len(input) == 1)
            """
        )

    with pytest.raises(Exception):
        execute_workflow(
            r"""
            [0]
            input: 'a.txt', 'b.txt'
            fail_if(len(input) == 2)
        """
        )


def test_delayed_fail_if():
    # test fail_if of killing another running substep
    st = time.time()
    with pytest.raises(Exception):
        execute_workflow(
            r"""
            import time

            [10]
            time.sleep(8)

            [20]
            input: None
            time.sleep(2)
            fail_if(True)
            """,
            config={"worker_procs": ["3"]},
        )

    assert (
        time.time() - st >= 8
    ), "Test test should fail only after step 10 is completed"


def test_delayed_fail_if_from_nested_workflow():
    # test fail_if of killing another running substep

    st = time.time()
    with pytest.raises(Exception):
        execute_workflow(
            r"""
            import time

            [default]
            sos_run('a')

            [a_10]
            time.sleep(8)

            [a_20]
            input: None
            time.sleep(2)
            fail_if(True)
        """,
            config={"worker_procs": ["3"]},
        )

    assert (
        time.time() - st >= 8
    ), "Test test should fail only after step 10 is completed"


def test_warn_if(temp_factory):
    """Test action fail if"""
    execute_workflow(
        r"""
        [0]
        warn_if(input is None, 'Expect to see a warning message')
        """,
        options={"run_mode": "dryrun"},
    )

    temp_factory("a.txt", "b.txt")
    execute_workflow(
        r"""
        [0]
        input: 'a.txt', 'b.txt'
        warn_if(len(_input) == 1)
        """,
        options={"run_mode": "dryrun"},
    )


def test_stop_if():
    """Test action stop_if"""
    execute_workflow(
        r"""
        [0: shared='result']
        rep = range(20)
        result = []
        input: for_each='rep'

        stop_if(_rep > 10)
        result.append(_rep)
        """
    )
    assert env.sos_dict["result"] == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]


def test_stop_if_1(clear_now_and_after):
    # stop_if should not be treated as error so the previously
    # generated output file will be removed
    clear_now_and_after("test_stop_if_0.txt", "test_stop_if_1.txt")

    execute_workflow(
        r"""
        [10]
        rep = range(2)
        input: for_each='rep'
        output: f'test_stop_if_{_rep}.txt'

        _output.touch()
        stop_if(_rep == 1, no_output=True)

        [20]
        assert(step_input.contains('test_stop_if_0.txt'))
        assert(not step_input.contains('test_stop_if_1.txt'))
        """
    )

    assert os.path.isfile("test_stop_if_0.txt")
    assert not (os.path.isfile("test_stop_if_1.txt"))


def test_stop_if_2(clear_now_and_after):
    # stop_if should not be treated as error so the previously
    # generated output file will not be removed
    clear_now_and_after("test_stop_if_0.txt", "test_stop_if_1.txt")

    execute_workflow(
        r"""
        [10]
        rep = range(2)
        input: for_each='rep'
        output: f'test_stop_if_{_rep}.txt'

        _output.touch()
        stop_if(_rep == 1)

        [20]
        assert(step_input.contains('test_stop_if_0.txt'))
        assert(step_input.contains('test_stop_if_1.txt'))
        """
    )
    assert os.path.isfile("test_stop_if_0.txt")
    assert os.path.isfile("test_stop_if_1.txt")


def test_skip_if():
    """Test action stop_if"""
    execute_workflow(
        r"""
        [0: shared='result']
        rep = range(20)
        result = []
        input: for_each='rep'

        skip_if(_rep > 10)
        result.append(_rep)
        """
    )
    assert env.sos_dict["result"] == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]


def test_skip_if_1(clear_now_and_after):
    clear_now_and_after([f"test_stop_if_{rep}.txt" for rep in range(2)])

    # stop_if should not be treated as error so the previously
    # generated output file will be removed

    execute_workflow(
        r"""
        [10]
        rep = range(2)
        input: for_each='rep'
        output: f'test_stop_if_{_rep}.txt'

        _output.touch()
        skip_if(_rep == 1)

        [20]
        assert(step_input.contains('test_stop_if_0.txt'))
        assert(not step_input.contains('test_stop_if_1.txt'))

        """
    )

    assert os.path.isfile("test_stop_if_0.txt")
    assert not os.path.isfile("test_stop_if_1.txt")


def test_done_if(clear_now_and_after):
    "Test action done_if"
    clear_now_and_after([f"test_done_if_{rep}.txt" for rep in range(2)])

    execute_workflow(
        r"""
        [10]
        rep = range(2)
        input: for_each='rep'
        output: f'test_done_if_{_rep}.txt'

        _output.touch()
        done_if(_rep == 1)

        [20]
        assert(step_input.contains('test_done_if_0.txt'))
        assert(step_input.contains('test_done_if_1.txt'))

        """
    )
    assert os.path.isfile("test_done_if_0.txt")
    assert os.path.isfile("test_done_if_1.txt")


def test_run():
    """Test action run"""
    execute_workflow(
        r"""
        [0]
        run:
        echo 'Echo'
        """
    )


@pytest.mark.skipif(
    sys.platform == "win32", reason="Under windows, echo 'Echo is perfectly OK"
)
def test_run_1():

    with pytest.raises(Exception):
        execute_workflow(
            r"""
            [0]
            run:
            echo 'Echo
            """
        )


def test_run_with_shebang():
    execute_workflow(
        r"""
        [0]
        run:
        #!/usr/bin/env python
        print('Echo')
        """
    )


@pytest.mark.skipif(not shutil.which("perl"), reason="Needs perl")
def test_perl():
    """Test action perl"""
    execute_workflow(
        r"""
        [0]
        perl:
        use strict;
        use warnings;

        print "hi NAME\n";
        """
    )


@pytest.mark.skipif(not shutil.which("ruby"), reason="Needs ruby")
def test_ruby():
    """Test action ruby"""
    execute_workflow(
        r"""
        [0]
        ruby:
        line1 = "Cats are smarter than dogs";
        line2 = "Dogs also like meat";

        if ( line1 =~ /Cats(.*)/ )
        puts "Line1 contains Cats"
        end
        if ( line2 =~ /Cats(.*)/ )
        puts "Line2 contains  Dogs"
        end
        """
    )


@pytest.mark.skipif(
    not with_network or "TRAVIS" in os.environ or "APPVEYOR" in os.environ,
    reason="Skip test because of no internet connection or in travis test",
)
def test_download(temp_factory, clear_now_and_after):
    """Test download of resources"""

    clear_now_and_after("tmp")
    temp_factory(dir="tmp")

    # test decompress tar.gz file
    execute_workflow(
        r"""
        [0]
        download(['http://bioinformatics.mdanderson.org/Software/VariantTools/repository/annoDB/CancerGeneCensus-20170912.DB.gz'],
        dest_dir='tmp', decompress=True)
        """
    )
    assert os.path.isfile("tmp/CancerGeneCensus-20170912.DB")

    #
    # testing the download of single file
    #
    execute_workflow(
        r"""
        [0]
        download: dest_file='tmp/refgene.ppp'
        http://bioinformatics.mdanderson.org/Software/VariantTools/repository/resource/refgene.pkl
        """
    )

    assert os.path.isfile("tmp/refgene.ppp")

    # test option dest_dir
    execute_workflow(
        r"""
        [0]
        download: dest_dir='tmp'
        http://bioinformatics.mdanderson.org/Software/VariantTools/repository/resource/refgene.pkl
        """
    )
    assert os.path.isfile("tmp/refgene.pkl")


@pytest.mark.skipif(
    not with_network or "TRAVIS" in os.environ or "APPVEYOR" in os.environ,
    reason="Skip test because of no internet connection or in travis test",
)
def test_download_missing_file(temp_factory, clear_now_and_after):

    clear_now_and_after("tmp")
    temp_factory(dir="tmp")

    with pytest.raises(Exception):
        execute_workflow(
            r"""
            [0]
            download: dest_dir='tmp', decompress=True, max_jobs=2
            http://bioinformatics.mdanderson.org/Software/VariantTools/repository/resource/non-existing.gz
            http://bioinformatics.mdanderson.org/Software/VariantTools/repository/annoDB/CancerGeneCensus-20170912.DB.gz
            http://bioinformatics.mdanderson.org/Software/VariantTools/repository/annoDB/CancerGeneCensus.ann
            http://bioinformatics.mdanderson.org/Software/VariantTools/repository/annoDB/DGV-hg38_20160831.ann
            """
        )

    assert os.path.isfile("tmp/CancerGeneCensus.ann")


@pytest.mark.skipif(
    not with_network or "TRAVIS" in os.environ or "APPVEYOR" in os.environ,
    reason="Skip test because of no internet connection or in travis test",
)
def test_download_large_file(temp_factory, clear_now_and_after):
    # test decompress tar.gz, .zip and .gz files

    clear_now_and_after("tmp")
    temp_factory(dir="tmp")

    execute_workflow(
        r"""
        [0]
        download: dest_dir='tmp', decompress=True
        http://bioinformatics.mdanderson.org/Software/VariantTools/repository/annoDB/CancerGeneCensus-20170912.DB.gz
        http://bioinformatics.mdanderson.org/Software/VariantTools/repository/annoDB/CancerGeneCensus.ann
        http://bioinformatics.mdanderson.org/Software/VariantTools/repository/annoDB/DGV-hg38_20160831.ann
        """
    )

    # run in build mode
    execute_workflow(
        r"""
        [0]
        download: dest_dir='tmp', decompress=True
        http://bioinformatics.mdanderson.org/Software/VariantTools/repository/annoDB/CancerGeneCensus-20170912.DB.gz
        http://bioinformatics.mdanderson.org/Software/VariantTools/repository/annoDB/CancerGeneCensus.ann
        http://bioinformatics.mdanderson.org/Software/VariantTools/repository/annoDB/DGV-hg38_20160831.ann
        """
    )


@pytest.mark.skipif(not shutil.which("pandoc"), reason="Needs pandoc")
def test_pandoc(clear_now_and_after):
    """Test action pandoc"""
    clear_now_and_after("report.md", "myreport.html")

    execute_workflow(
        r"""
        [10]

        report: output='report.md'
        ## Some random figure

        Generated by matplotlib


        [100]
        # generate report
        output: 'myreport.html'
        pandoc(input='report.md', output=_output[0])
        """
    )

    assert os.path.isfile("myreport.html")


@pytest.mark.skipif(not shutil.which("pandoc"), reason="Needs pandoc")
def test_pandoc_1(clear_now_and_after):
    clear_now_and_after("a.md", "myreport.html")

    execute_workflow(
        r"""
        [10]
        report: output='a.md'
        ## Some random figure

        Generated by matplotlib


        [100]
        # generate report
        output: 'myreport.html'
        pandoc(input='a.md', output=_output[0])
        """
    )

    assert os.path.isfile("myreport.html")


@pytest.mark.skipif(not shutil.which("pandoc"), reason="Needs pandoc")
def test_pandoc_2(clear_now_and_after):
    clear_now_and_after("a.md", "myreport.html")
    #
    # another case is no output
    execute_workflow(
        r"""
        [10]
        report: output='a.md'
        ## Some random figure

        Generated by matplotlib


        [100]
        # generate report
        pandoc(input='a.md')
        """
    )


@pytest.mark.skipif(not shutil.which("pandoc"), reason="Needs pandoc")
def test_pandoc_3(clear_now_and_after):
    # test acceptance of a list of input filenames
    clear_now_and_after("default_10.md", "default_20.md", "output.html")

    execute_workflow(
        r"""
        [10]
        report: output='default_10.md'
        A_10

        [20]
        report: output='default_20.md'
        A_20

        [100]
        # generate report
        pandoc(input=['default_10.md', 'default_20.md'], output='output.html')
        """
    )

    for f in ["default_10.md", "default_20.md", "output.html"]:
        assert file_target(f).target_exists()


def test_report(clear_now_and_after):
    """Test action report"""
    clear_now_and_after("report.txt")

    script = r"""
        [A]
        parameter: num=5
        report: output='report.txt', expand=True
        touch {num}.txt

        """
    # run twice
    execute_workflow(script, args=["--num", "7"])
    execute_workflow(script, args=["--num", "5"])

    with open("report.txt") as report:
        assert report.read() == "touch 5.txt\n\n"


def test_report_1(clear_now_and_after):
    """Test action report"""
    clear_now_and_after("report.txt")

    execute_workflow(
        r"""
        [A]
        report: output='report.txt', expand=True
        {step_name}

        [A_10]
        report: output='report.txt', expand=True
        {step_name}
        """
    )
    with open("report.txt") as report:
        assert report.read() == "A_10\n\n"


def test_report_2(clear_now_and_after):
    """Test action report"""
    clear_now_and_after("a.txt", "out.txt")

    execute_workflow(
        r"""
        [A_1]
        run: output='a.txt'
            echo something > a.txt

        report(input='a.txt', output='out.txt')
        """
    )

    for name in ("a.txt", "out.txt"):
        with open(name) as report:
            assert report.read().strip() == "something"


def test_report_3(clear_now_and_after):
    """Test action report"""
    clear_now_and_after("report.txt")
    #
    execute_workflow(
        r"""
        [A_1]
        run: output='a.txt'
        echo something > a.txt

        [A_2]
        run: output='b.txt'
        echo something else > b.txt

        [A_3]
        report(input=['a.txt', 'b.txt'], output='out.txt')
        """
    )
    for name in ("a.txt", "b.txt", "out.txt"):
        assert file_target(name).target_exists()


def test_report_4(clear_now_and_after):
    """Test action report"""
    clear_now_and_after("a.txt")

    # test report to other types of output: path
    execute_workflow(
        r"""
        [A_1]
        report: output=path('a.txt')
        something
        """
    )


def test_report_5(clear_now_and_after):
    """Test action report"""
    clear_now_and_after("a.txt")

    execute_workflow(
        r"""
        [A_1]
        output: 'a.txt'
        report: output=_output[0]
        something
        """
    )


def test_report_6(clear_now_and_after):
    """Test action report"""
    clear_now_and_after("a.txt")

    # test report to other types of output: sos_targets
    execute_workflow(
        r"""
        [A_1]
        output: 'a.txt'
        report: output=_output
        something
        """
    )


def test_report_7(clear_now_and_after):
    """Test action report"""
    clear_now_and_after("a.txt")

    with pytest.raises(Exception):
        execute_workflow(
            r"""
            [A_1]
            output: 'a.txt', 'b.txt'
            report: output=_output
            something
            """
        )


def test_option_workdir(temp_factory):
    """Test option workdir of tasks"""
    temp_factory(dir="temp_wdr")

    with open(os.path.join("temp_wdr", "a.txt"), "w") as tmp:
        tmp.write("hello")

    execute_workflow(
        r"""
        [A_1]
        run: workdir='temp_wdr'
        cp -f a.txt a2.txt
        """
    )

    assert file_target(os.path.join("temp_wdr", "a2.txt")).target_exists()
    with open(os.path.join("temp_wdr", "a.txt")) as tmp:
        assert "hello" == tmp.read()


def test_action_script():
    """Test action script"""
    execute_workflow(
        r"""
        [A_1]
        script: interpreter='python'
        with open('something.txt', 'w') as tmp:
            tmp.write('something')
        """
    )

    assert file_target("something.txt").target_exists()
    with open("something.txt") as tmp:
        assert "something" == tmp.read()


def test_regenerate_report(clear_now_and_after):
    """Testing the regeneration of report once is needed. The problem
    here is the 'input' parameter of report."""
    clear_now_and_after("a1.md", "a2.md", "out.md")
    script = r"""
        [A_1]
        output: 'a1.txt', 'a1.md'
        run:
        echo 'a1' >> a1.txt

        report: output='a1.md'
        a1

        [A_2]
        output: 'a2.txt', 'a2.md'
        run:
        echo 'a2' >> a2.txt
        report: output='a2.md'
        a2

        [A_3]
        input: 'a1.md', 'a2.md'
        output: 'out.md'
        report:     input=['a1.md', 'a2.md'], output='out.md'
        """
    execute_workflow(script)
    with open("a1.md") as a:
        assert a.read() == "a1\n\n"
    with open("a2.md") as a:
        assert a.read() == "a2\n\n"
    with open("out.md") as a:
        assert a.read() == "a1\n\na2\n\n"

    clear_now_and_after("a1.md", "a2.md", "out.md")
    execute_workflow(script)


def test_active_action_option(temp_factory):
    """Test the active option of actions"""
    # disallow
    with pytest.raises(Exception):
        SoS_Script(
            """
[1]
rep = range(5)
input: for_each = 'rep'
# ff should change and be usable inside run
ff = f"{_rep}.txt"
run:  expand=True, active=1,2
echo {ff}
touch temp/{ff}
"""
        )
    #
    for active, result in [
        ("0", ["temp/0.txt"]),
        ("-1", ["temp/4.txt"]),
        ("(1,2)", ["temp/1.txt", "temp/2.txt"]),
        ("[2,3]", ["temp/2.txt", "temp/3.txt"]),
        ("(0,2,4)", ["temp/0.txt", "temp/2.txt", "temp/4.txt"]),
        ("slice(1,None)", ["temp/1.txt", "temp/2.txt", "temp/3.txt", "temp/4.txt"]),
        ("slice(1,-2)", ["temp/1.txt", "temp/2.txt"]),
        ("slice(None,None,2)", ["temp/0.txt", "temp/2.txt", "temp/4.txt"]),
        (
            "True",
            ["temp/0.txt", "temp/1.txt", "temp/2.txt", "temp/3.txt", "temp/4.txt"],
        ),
        ("False", []),
    ]:
        temp_factory(dir="temp")
        # test first iteration
        execute_workflow(
            (
                """
            [1]
            rep = range(5)
            input: for_each = 'rep'
            # ff should change and be usable inside run
            ff = f"{_rep}.txt"
            run:  expand=True, active=%s
            echo {ff}
            touch temp/{ff}
            """
                % active
            ).replace("/", os.sep),
            options={"sig_mode": "force"},
        )

        files = list(glob.glob(os.path.join("temp", "*.txt")))
        assert sorted(files) == sorted([x.replace("/", os.sep) for x in result])


def test_action_option_template(clear_now_and_after):
    clear_now_and_after("template_output.txt")

    execute_workflow(
        """
        run:  template='cat {filename}'
            echo 'whatever' > template_output.txt
        """
    )
    assert not os.path.isfile("template_output.txt")


def test_action_option_template_name(config_factory):
    cfg = config_factory(
        """\
        action_templates:
            cat: |
                cat {filename}

        """
    )
    execute_workflow(
        """
        run:  template_name='cat'
            echo 'whatever' > template_output.txt
        """,
        args=["-c", cfg],
    )
    assert not os.path.isfile("template_output.txt")



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.
import tempfile
import os
import sys
import zmq
import time
import threading
import uuid
from collections import defaultdict
from .utils import env, ProcessKilled, get_localhost_ip, get_open_files_and_connections
from .signatures import StepSignatures, WorkflowSignatures
from .messages import encode_msg, decode_msg

EVENT_MAP = {}
for name in ('PUSH', 'PULL', 'PAIR', 'REQ', 'REP'):
    EVENT_MAP[getattr(zmq, name)] = name

g_sockets = set()


def create_socket(context, socket_type, desc=''):
    socket = context.socket(socket_type)
    g_sockets.add(socket.fd)
    if 'CONTROLLER' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
            'SOS_DEBUG']:
        env.log_to_file(
            'CONTROLLER',
            f'{os.getpid()} {desc}: new socket of type {EVENT_MAP.get(socket_type, "UNKNOWN")} with handler {socket.fd} ({len(g_sockets)} total)'
        )
    return socket


def close_socket(socket, desc='', now=False):
    if socket is None:
        return
    g_sockets.remove(socket.fd)
    if 'CONTROLLER' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
            'SOS_DEBUG']:
        env.log_to_file(
            'CONTROLLER',
            f'{os.getpid()} {desc}: closes socket with handler {socket.fd} ({len(g_sockets)} left)'
        )
    if now:
        socket.LINGER = 0
    socket.close()
    return socket


def zmq_term(context):
    #
    # the following is only valid when multiprocessing is started with the spawn method
    # otherwise g_sockets will contain sockets from the parental process

    # if g_sockets:
    # env.logger.warning(f'{os.getpid()} terminting zmq with {len(g_sockets)} unclosed sockets: {g_sockets}')
    context.term()


def send_message_to_controller(msg):
    if env.master_push_socket is None:
        env.master_push_socket = create_socket(env.zmq_context, zmq.PUSH,
                                               'master push')
        env.master_push_socket.connect(env.config["sockets"]["master_push"])
    env.master_push_socket.send(encode_msg(msg))


def request_answer_from_controller(msg):
    if env.master_request_socket is None:
        env.master_request_socket = create_socket(env.zmq_context, zmq.REQ,
                                                  'master request')
        env.master_request_socket.connect(
            env.config["sockets"]["master_request"])
    env.master_request_socket.send(encode_msg(msg))
    return decode_msg(env.master_request_socket.recv())


def connect_controllers(context=None):
    if not context:
        if 'CONTROLLER' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                'SOS_DEBUG']:
            env.log_to_file('CONTROLLER', f'create context at {os.getpid()}')
        context = zmq.Context()

    if 'CONTROLLER' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
            'SOS_DEBUG']:
        env.log_to_file('CONTROLLER', f'Connecting sockets from {os.getpid()}')

    env.master_push_socket = None
    env.master_request_socket = None

    # if this instance of sos is being tapped. It should connect to a few sockets
    #
    if env.config['exec_mode'] == 'slave':
        env.tapping_logging_socket = create_socket(context, zmq.PUSH)
        env.tapping_logging_socket.connect(
            env.config["sockets"]["tapping_logging"])
        # change logging to socket
        env.set_socket_logger(env.tapping_logging_socket)

    # master also need to update task status from interactive runner.
    if env.config['exec_mode'] in ('master', 'slave'):
        env.tapping_listener_socket = create_socket(context, zmq.PUSH)
        env.tapping_listener_socket.connect(
            env.config["sockets"]["tapping_listener"])

    return context


def disconnect_controllers(context=None):
    close_socket(env.master_push_socket, now=True)
    close_socket(env.master_request_socket, now=True)

    if env.config['exec_mode'] == 'slave':
        close_socket(env.tapping_logging_socket, now=True)
        env.set_socket_logger(None)

    if env.config['exec_mode'] in ('master', 'slave'):
        close_socket(env.tapping_listener_socket, now=True)

    if 'CONTROLLER' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
            'SOS_DEBUG']:
        env.log_to_file('CONTROLLER',
                        f'Disconnecting sockets from {os.getpid()}')

    if context:
        if 'CONTROLLER' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                'SOS_DEBUG']:
            env.log_to_file('CONTROLLER', f'terminate context at {os.getpid()}')
        zmq_term(context)


class DotProgressBar:

    def __init__(self, context, interval=1):
        self.context = context
        self.interval = interval * 1000

        self._subprogressbar_size = 25
        self._substep_last_updated = time.time()

        self.stop_event = threading.Event()
        self._substep_cnt = 0

        # broker to handle the execution of substeps
        self.progress_push_socket = create_socket(self.context, zmq.PUSH,
                                                  'progress push')
        self.progress_port = self.progress_push_socket.bind_to_random_port(
            'tcp://127.0.0.1')

        self.thread = threading.Thread(target=self.run)
        self.thread.start()

    def run(self):
        progress_pull_socket = create_socket(self.context, zmq.PULL,
                                             'progress pull')
        progress_pull_socket.connect(f'tcp://127.0.0.1:{self.progress_port}')

        # leading progress bar
        sys.stderr.write('\033[32m[\033[0m')
        sys.stderr.flush()

        _pulse_cnt = 0
        while True:
            # no new message, add pulse
            if self.stop_event.is_set():
                return
            if not progress_pull_socket.poll(self.interval):
                if _pulse_cnt == 10:
                    sys.stderr.write('\b \b' * _pulse_cnt)
                    _pulse_cnt = 0
                else:
                    sys.stderr.write('\033[97m.\033[0m')
                    _pulse_cnt += 1
                sys.stderr.flush()
            else:
                msg = progress_pull_socket.recv().decode()
                # print update message
                sys.stderr.write('\b \b' * _pulse_cnt + msg)
                _pulse_cnt = 0
                sys.stderr.flush()

    def update(self, prog_type, status=None):
        if prog_type == 'substep_ignored':
            if time.time() - self._substep_last_updated < 1:
                return
            if self._substep_cnt == self._subprogressbar_size:
                update_str = '\b \b' * self._substep_cnt + '\033[90m.\033[0m'
                self._substep_cnt = 0
            else:
                update_str = '\033[90m.\033[0m'
            self._substep_cnt += 1
            self._substep_last_updated = time.time()
        elif prog_type == 'substep_completed':
            if time.time() - self._substep_last_updated < 1:
                return
            if self._substep_cnt == self._subprogressbar_size:
                update_str = '\b \b' * self._substep_cnt + '\033[32m.\033[0m'
                self._substep_cnt = 0
            else:
                update_str = '\033[32m.\033[0m'
            self._substep_cnt += 1
            self._substep_last_updated = time.time()
        elif prog_type == 'step_completed':
            update_str = '\b \b' * self._substep_cnt
            self._substep_cnt = 0
            if status == 1:  # completed
                update_str += '\033[32m#\033[0m'
            elif status == 0:  # completed
                update_str += '\033[90m#\033[0m'
            elif status > 0:  # in the middle
                update_str += '\033[36m#\033[0m'
            else:  # untracked (no signature)
                update_str += '\033[33m#\033[0m'
        elif prog_type == 'done':
            update_str = '\b \b' * self._substep_cnt + f'\033[32m]\033[0m {status}\n'
            self._substep_cnt = 0

        self.progress_push_socket.send(update_str.encode())

    def done(self, msg):
        self.update('done', msg)
        self.stop_event.set()


class Controller(threading.Thread):
    '''This controller is used by both sos and sos-notebook, and there
    can be two controllers one as a slave (sos) and one as a master
    (notebook). We shared the same code base because step executors need
    need to talk to the same controller (signature, controller etc) when
    they are executed in sos or sos notebook.
    '''

    def __init__(self, ready, kernel=None):
        threading.Thread.__init__(self)
        #self.daemon = True

        self.step_signatures = StepSignatures()
        self.workflow_signatures = WorkflowSignatures()

        self.tapping_controller_socket = None
        self.tapping_listener_socket = None

        self.ready = ready
        self.kernel = kernel
        # number of active running master processes
        self._nprocs = 0

        self._completed = defaultdict(int)
        self._ignored = defaultdict(int)

        # completed steps
        self._completed_steps = {}

        # substep workers
        self.workers = None

        # available resources
        self._resources = {}

        # temporary files
        self._tempfiles = {}

        # self.event_map = {}
        # for name in dir(zmq):
        #     if name.startswith('EVENT_'):
        #         value = getattr(zmq, name)
        #          self.event_map[value] = name
        self.console_logger = None

    def handle_master_push_msg(self, msg):
        try:
            if msg[0] in ('substep', 'step', 'workflow', 'task'):
                # cache the request, route to first available worker
                self.workers.add_request(msg[0], msg[1])
            elif msg[0] == 'nprocs':
                if 'CONTROLLER' in env.config[
                        'SOS_DEBUG'] or 'ALL' in env.config['SOS_DEBUG']:
                    env.log_to_file('CONTROLLER',
                                    f'Active running process set to {msg[1]}')
                self._nprocs = msg[1]
            elif msg[0] == 'progress':
                if msg[1] == 'substep_ignored':
                    self._ignored[msg[2]] += 1
                elif msg[1] == 'substep_completed':
                    self._completed[msg[2]] += 1
                elif msg[1] == 'step_completed':
                    self._completed_steps[msg[3]] = msg[4]
                if env.verbosity == 0:
                    # update progress bar
                    self._progress_bar.update(msg[1],
                                              msg[2] if len(msg) > 2 else None)
            elif msg[0] == 'workflow_sig':
                self.workflow_signatures.write(*msg[1:])
            elif msg[0] == 'step_sig':
                self.step_signatures.set(*msg[1:])
            elif msg[0] == 'commit_sig':
                self.workflow_signatures.commit()
                self.step_signatures.commit()
            elif msg[0] == 'resource':
                if msg[1] == 'docker_image':
                    if msg[2] in ('available', 'unavailable'):
                        if 'docker_image' not in self._resources:
                            self._resources['docker_image'] = {}
                        self._resources['docker_image'][msg[3]] = msg[2]
                    else:
                        raise ValueError(
                            f'SoS currently only understand "available" or "unavailable" messages for docker resource.'
                        )
                elif msg[1] == 'singularity_image':
                    if msg[2] in ('available', 'unavailable'):
                        if 'singularity_image' not in self._resources:
                            self._resources['singularity_image'] = {}
                        self._resources['singularity_image'][msg[3]] = msg[2]
                    else:
                        raise ValueError(
                            f'SoS currently only understand "available" or "unavailable" messages for docker resource.'
                        )
                else:
                    raise ValueError(
                        f'SoS currently does not handle resource of {msg[1]} type'
                    )

            else:
                env.logger.warning(f'Unknown message passed {msg}')
        except Exception as e:
            env.logger.warning(
                f'Failed to handle master push message {msg}: {e}')

    def handle_master_request_msg(self, msg):
        try:
            # make sure all records have been saved before returning information
            while True:
                if self.master_push_socket.poll(0):
                    self.handle_master_push_msg(
                        decode_msg(self.master_push_socket.recv()))
                else:
                    break
            if msg[0] == 'workflow_sig':
                if msg[1] == 'clear':
                    self.workflow_signatures.clear()
                    self.master_request_socket.send(encode_msg('ok'))
                elif msg[1] == 'placeholders':
                    self.master_request_socket.send(
                        encode_msg(
                            self.workflow_signatures.placeholders(msg[2])))
                elif msg[1] == 'records':
                    self.master_request_socket.send(
                        encode_msg(self.workflow_signatures.records(msg[2])))
                else:
                    env.logger.warning(f'Unknown signature request {msg}')
            elif msg[0] == 'step_sig':
                if msg[1] == 'get':
                    self.master_request_socket.send(
                        encode_msg(self.step_signatures.get(*msg[2:])))
                else:
                    env.logger.warning(f'Unknown signature request {msg}')
            elif msg[0] == 'sos_tempfile':
                path, name, suffix, prefix, dir = msg[1:]
                if path is not None:
                    if path not in self._tempfiles.values():
                        if not name:
                            name = uuid.uuid4().hex
                        self._tempfiles[name] = path
                    self.master_request_socket.send(encode_msg(path))
                elif name is not None and name in self._tempfiles:
                    self.master_request_socket.send(
                        encode_msg(self._tempfiles[name]))
                else:
                    #
                    basename = uuid.uuid4().hex
                    filename = basename
                    if name is None:
                        name = basename
                    if suffix is not None:
                        filename = filename + suffix
                    if prefix is not None:
                        filename = prefix + filename
                    if dir is not None:
                        filename = os.path.join(dir, filename)
                    else:
                        filename = os.path.join(tempfile.gettempdir(), filename)
                    #
                    self._tempfiles[name] = filename
                    self.master_request_socket.send(encode_msg(filename))
            elif msg[0] == 'nprocs':
                self.master_request_socket.send(encode_msg(self._nprocs))
            elif msg[0] == 'sos_step':
                self.master_request_socket.send(
                    encode_msg(msg[1] in self._completed_steps or msg[1] in [
                        x.rsplit('_', 1)[0]
                        for x in self._completed_steps.keys()
                    ]))
            elif msg[0] == 'step_output':
                step_name = msg[1]
                if step_name in self._completed_steps:
                    self.master_request_socket.send(
                        encode_msg(self._completed_steps[step_name]))
                else:
                    # now, step_name might actually be a workflow name, in which
                    # case we need to return the last step of the workflow
                    steps = sorted([
                        x for x in self._completed_steps.keys()
                        if x.rsplit('_', 1)[0] == step_name
                    ])
                    self.master_request_socket.send(
                        encode_msg(self._completed_steps[steps[-1]]
                                   if steps else None))
            elif msg[0] == 'named_output':
                name = msg[1]
                found = False
                for step_output in self._completed_steps.values():
                    if name in step_output.labels:
                        found = True
                        self.master_request_socket.send(
                            encode_msg(step_output[name]))
                        break
                if not found:
                    self.master_request_socket.send(encode_msg(None))
            elif msg[0] == 'worker_available':
                self.master_request_socket.send(
                    encode_msg(self.workers.worker_available(msg[1], msg[2:])))
            elif msg[0] == 'resource':
                if msg[1] == 'docker_image':
                    if msg[2] == 'request':
                        if 'docker_image' not in self._resources:
                            self._resources['docker_image'] = {}
                        if msg[3] in self._resources['docker_image']:
                            self.master_request_socket.send(
                                encode_msg(
                                    self._resources['docker_image'][msg[3]]))
                        else:
                            self._resources['docker_image'][msg[3]] = 'pending'
                            self.master_request_socket.send(
                                encode_msg('help yourself'))
                    else:
                        raise ValueError(
                            f'SoS currently only accept request inquiry for docker resource'
                        )
                elif msg[1] == 'singularity_image':
                    if msg[2] == 'request':
                        if 'singularity_image' not in self._resources:
                            self._resources['singularity_image'] = {}
                        if msg[3] in self._resources['singularity_image']:
                            self.master_request_socket.send(
                                encode_msg(self._resources['singularity_image'][
                                    msg[3]]))
                        else:
                            self._resources['singularity_image'][
                                msg[3]] = 'pending'
                            self.master_request_socket.send(
                                encode_msg('help yourself'))
                    else:
                        raise ValueError(
                            f'SoS currently only accept request inquiry for docker resource'
                        )
                else:
                    raise ValueError(
                        f'SoS currently does not handle resource of {msg[1]} type'
                    )
            elif msg[0] == 'done':
                # handle all ctl_push_msgs #1062
                while True:
                    if self.master_push_socket.poll(0):
                        self.handle_master_push_msg(
                            decode_msg(self.master_push_socket.recv()))
                    else:
                        break

                # handle all push request from logging
                if env.config['exec_mode'] in ('master', 'both'):
                    while True:
                        if self.tapping_logging_socket.poll(0):
                            self.handle_tapping_logging_msg(
                                self.tapping_logging_socket.recv_multipart())
                        else:
                            break

                if env.verbosity == 0:
                    num_steps = len(
                        set(self._completed.keys())
                        | set(self._ignored.keys()))
                    num_completed = sum(self._completed.values())
                    num_ignored = sum(self._ignored.values())
                    completed_text = f'{num_completed} job{"s" if num_completed > 1 else ""} completed' if num_completed else ''
                    ignored_text = f'{num_ignored} job{"s" if num_ignored > 1 else ""} ignored' if num_ignored else ''
                    steps_text = f'{num_steps} step{"s" if num_steps > 1 else ""} processed'
                    succ = '' if msg[1] else 'Failed with '
                    self._progress_bar.done(
                        f'{succ}{steps_text} ({completed_text}{", " if num_completed and num_ignored else ""}{ignored_text})'
                    )

                self.master_request_socket.send(encode_msg('bye'))

                return False
            else:
                raise RuntimeError(f'Unrecognized request {msg}')
            return True
        except Exception as e:
            env.logger.warning(f'Failed to respond controller {msg}: {e}')
            self.master_request_socket.send(encode_msg(None))

    def handle_worker_backend_msg(self, msg):
        # msg should be a port number from the worker
        self.workers.process_request(msg[0], msg[1:])

    def handle_tapping_logging_msg(self, msg):
        if env.config['exec_mode'] == 'both':
            print(' '.join(x.decode() for x in msg))
        elif msg[0] == b'ERROR':
            env.logger.error(msg[1].decode())
        elif msg[0] == b'WARNING':
            env.logger.warning(msg[1].decode())
        elif msg[0] == b'INFO':
            env.logger.info(msg[1].decode())
        elif msg[0] == b'DEBUG':
            env.logger.debug(msg[1].decode())
        elif msg[0] == b'TRACE':
            if 'CONTROLLER' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                    'SOS_DEBUG']:
                env.log_to_file('CONTROLLER', msg[1].decode())
        elif msg[0] == b'PRINT':
            env.logger.print(*[x.decode() for x in msg[1:]])
        else:
            print(' '.join(x.decode() for x in msg))

    def handle_tapping_listener_msg(self, msg):
        try:
            #env.log_to_file(f'listener got {msg}')
            self.kernel.send_frontend_msg(msg['msg_type'], msg['data'])
        except Exception as e:
            env.log_to_file(
                f'Failed to handle tapping listerner message {msg}: {e}')

    def handle_tapping_controller_msg(self, msg):
        self.tapping_controller_socket.send(b'ok')

    def run(self):
        # there are two sockets
        #
        self.context = zmq.Context.instance()
        local_ip = get_localhost_ip()

        if 'CONTROLLER' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                'SOS_DEBUG']:
            env.log_to_file('CONTROLLER', f'controller started {os.getpid()}')

        if 'sockets' not in env.config:
            env.config['sockets'] = {}

        self.master_push_socket = create_socket(self.context, zmq.PULL,
                                                'controller master_pull')
        master_push_port = self.master_push_socket.bind_to_random_port(
            f'tcp://{local_ip}')
        env.config['sockets'][
            'master_push'] = f'tcp://{local_ip}:{master_push_port}'

        self.master_request_socket = create_socket(self.context, zmq.REP,
                                                   'controller master_request')
        master_request_port = self.master_request_socket.bind_to_random_port(
            f'tcp://{local_ip}')
        env.config['sockets'][
            'master_request'] = f'tcp://{local_ip}:{master_request_port}'

        # broker to handle the execution of substeps
        self.worker_backend_socket = create_socket(
            self.context, zmq.REP, 'controller backend rep')  # ROUTER
        # we assume the router is always on local host, but we will use a non-localhost
        # IP so that others can connect to it.
        worker_port = self.worker_backend_socket.bind_to_random_port(
            f'tcp://{local_ip}')
        env.config['sockets'][
            'worker_backend'] = f'tcp://{local_ip}:{worker_port}'

        # tapping
        if env.config['exec_mode'] == 'master':
            self.tapping_logging_socket = create_socket(self.context, zmq.PULL)
            tapping_logging_port = self.tapping_logging_socket.bind_to_random_port(
                f'tcp://{local_ip}')
            env.config['sockets'][
                'tapping_logging'] = f'tcp://{local_ip}:{tapping_logging_port}'

            self.tapping_listener_socket = create_socket(self.context, zmq.PULL)
            tapping_listener_port = self.tapping_listener_socket.bind_to_random_port(
                f'tcp://{local_ip}')
            env.config['sockets'][
                'tapping_listener'] = f'tcp://{local_ip}:{tapping_listener_port}'

            self.tapping_controller_socket = create_socket(
                self.context, zmq.PUSH)
            tapping_controller_port = self.tapping_controller_socket.bind_to_random_port(
                f'tcp://{local_ip}')
            env.config['sockets'][
                'tapping_controller'] = f'tcp://{local_ip}:{tapping_controller_port}'

        if env.config['exec_mode'] == 'slave':
            self.tapping_controller_socket = create_socket(
                self.context, zmq.PULL)
            self.tapping_controller_socket.connect(
                env.config["sockets"]["tapping_controller"])

        #monitor_socket = self.master_request_socket.get_monitor_socket()
        # tell others that the sockets are ready
        self.ready.set()

        # create a manager
        from .workers import WorkerManager
        self.workers = WorkerManager(env.config['worker_procs'],
                                     self.worker_backend_socket)

        # Process messages from receiver and controller
        poller = zmq.Poller()
        poller.register(self.master_push_socket, zmq.POLLIN)
        poller.register(self.master_request_socket, zmq.POLLIN)
        poller.register(self.worker_backend_socket, zmq.POLLIN)
        if env.config['exec_mode'] == 'master':
            poller.register(self.tapping_logging_socket, zmq.POLLIN)
            poller.register(self.tapping_listener_socket, zmq.POLLIN)
        if env.config['exec_mode'] == 'slave':
            poller.register(self.tapping_controller_socket, zmq.POLLIN)

        #poller.register(monitor_socket, zmq.POLLIN)
        if env.verbosity == 0:
            # leading progress bar
            self._progress_bar = DotProgressBar(self.context)

        try:
            while True:
                if 'OPENFILES' in env.config[
                        'SOS_DEBUG'] or 'ALL' in env.config['SOS_DEBUG']:
                    ofc = get_open_files_and_connections(os.getpid())
                    env.log_to_file(
                        'OPENFILES',
                        f'CONTROLLER {os.getpid()} has {len(ofc["open_files"])} open files and {len(ofc["connections"])} connections.'
                    )
                while True:
                    socks = dict(poller.poll(1000))
                    if socks:
                        break
                    # if the last worker has been pending for more than 5
                    # seconds, kill it. It is also possible that some others are killed
                    # by external process.
                    self.workers.check_workers()

                if self.master_push_socket in socks:
                    while True:
                        if self.master_push_socket.poll(0):
                            self.handle_master_push_msg(
                                decode_msg(self.master_push_socket.recv()))
                        else:
                            break

                if self.master_request_socket in socks:
                    if not self.handle_master_request_msg(
                            decode_msg(self.master_request_socket.recv())):
                        break

                if self.worker_backend_socket in socks:
                    while True:
                        if self.worker_backend_socket.poll(0):
                            self.handle_worker_backend_msg(
                                decode_msg(self.worker_backend_socket.recv()))
                        else:
                            break

                if env.config['exec_mode'] == 'master':
                    if self.tapping_logging_socket in socks:
                        self.handle_tapping_logging_msg(
                            self.tapping_logging_socket.recv_multipart())
                    if self.tapping_listener_socket in socks:
                        self.handle_tapping_listener_msg(
                            decode_msg(self.tapping_listener_socket.recv()))

                if env.config['exec_mode'] == 'slave':
                    if self.tapping_controller_socket in socks:
                        self.handle_tapping_controller_msg(
                            decode_msg(self.tapping_controller_socket.recv()))

                # if monitor_socket in socks:
                #     evt = recv_monitor_message(monitor_socket)
                #     if evt['event'] == zmq.EVENT_ACCEPTED:
                #         self._num_clients += 1
                #     elif evt['event'] == zmq.EVENT_DISCONNECTED:
                #         self._num_clients -= 1
        except ProcessKilled as e:
            env.logger.error(str(e))
            os._exit(1)
        except Exception as e:
            sys.stderr.write(f'{env.config["exec_mode"]} get an error {e}')
            return
        finally:
            # remove temporary files
            for name, tempfile in self._tempfiles.items():
                try:
                    if os.path.isfile(tempfile):
                        os.remove(tempfile)
                except Exception as e:
                    env.logger.warning(
                        f'Failed to tempfile associated with name {name}')
            # kill all workers
            self.workers.kill_all()

            # close all databses
            self.step_signatures.close()
            self.workflow_signatures.close()

            poller.unregister(self.master_push_socket)
            poller.unregister(self.master_request_socket)
            poller.unregister(self.worker_backend_socket)
            if env.config['exec_mode'] == 'master':
                poller.unregister(self.tapping_logging_socket)
                poller.unregister(self.tapping_listener_socket)
            if env.config['exec_mode'] == 'slave':
                poller.unregister(self.tapping_controller_socket)

            close_socket(self.master_push_socket, now=True)
            close_socket(self.master_request_socket, now=True)
            close_socket(self.worker_backend_socket, now=True)

            if env.config['exec_mode'] == 'master':
                close_socket(self.tapping_logging_socket, now=True)
                close_socket(self.tapping_listener_socket, now=True)
            # both master and slave has it
            if env.config['exec_mode'] in ('master', 'slave'):
                close_socket(self.tapping_controller_socket, now=True)

            if 'CONTROLLER' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                    'SOS_DEBUG']:
                env.log_to_file('CONTROLLER',
                                f'controller stopped {os.getpid()}')



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import ast
import sys
import copy
import pickle
import contextlib

from typing import Any, Dict, Optional, Set

from .utils import env, as_fstring, load_config_files, pickleable, ArgumentError
from ._version import __version__


def interpolate(text, global_dict=None, local_dict=None):
    '''Evaluate expressions in `text` '''
    # step 1, make it a f-string (add quotation marks and f
    # step 2, evaluate as a string
    try:
        return eval(as_fstring(text), global_dict, local_dict)
    except Exception as e:
        raise ValueError(f'Failed to interpolate {text}: {e}')


def cfg_interpolate(text, local_dict={}):
    # handle nested interpolate ...
    # we need to avoid containing our CONFIG with all the modules created by eval
    cfg_dict = copy.deepcopy(env.sos_dict.get('CONFIG', {}))
    if 'os.environ' in text:
        exec('import os', cfg_dict)
    while True:
        res = interpolate(text, cfg_dict, local_dict)
        if res == text:
            break
        else:
            text = res
    return res


def get_config(*args, **kwargs):
    '''
    Obtain configuration from sos configuration files, by default with variable intepolation.

    The basic way to use this function is to do

       get_config('key')

    which gets the value associated with configuration "key". If the value is a dictionary, you
    can go into the dictionary with any of the following

       get_config('key', 'key1')
       get_config(['key', 'key1'])
       get_config('key.key1')

    The expressions are by default interpolated with variables defined in the order of,

    1. specified from get_config
    2. the same dictionary in which the expression is evaluated
    3. the root of the configuration dictionary

    You can specify variables as a dictionary

        get_config('a', {'val': 5})

    if the key does not conflict with other keywords, you can specify directly as keyword arguments

        get_config('a', val=5)

    The funciton accepts the following keyword parameters:

      default: default value to return if the key is not found. Default to None.
      raw: return raw values without interpolation.
      raw_keys: keys that are not expanded
      expand_keys: keys that are expanded
      allowed_keys: retrieve only specified keys from a dicitonary.
      exclude_keys: exclude specified keys
      expected_type: expected type or types
    '''
    default = kwargs.get('default', None)
    allowed_keys = kwargs.get('allowed_keys', None)
    excluded_keys = kwargs.get('excluded_keys', None)
    raw_keys = kwargs.get('raw_keys', None)
    expand_keys = kwargs.get('expand_keys', None)
    raw = kwargs.get('raw', False)
    expected_type = kwargs.get('expected_type', None)

    #
    keys = []
    custom_dict = {}
    for arg in args:
        if isinstance(arg, str):
            keys.append(arg)
        elif isinstance(arg, (list, tuple)):
            keys.extend(list(arg))
        elif isinstance(arg, dict):
            custom_dict.update(arg)
        else:
            raise ValueError(f'Unacceptable parameter {arg} for get_config.')
    #
    custom_dict.update({
        x: y
        for x, y in kwargs.items()
        if x not in ('default', 'allowed_keys', 'excluded_keys', 'raw_keys',
                     'expand_keys', 'raw')
    })
    #
    local_dict = {}
    val = env.sos_dict.get('CONFIG', {})
    while True:
        if not keys:
            break
        if not isinstance(val, dict):
            raise ValueError(
                f'A dictionary is expected to get item with key {".".join(keys)}: {val} obtained.'
            )
        local_dict = val
        key = keys[0]
        if key in val:
            keys.pop(0)
            val = val[key]
            continue
        elif '.' in key:
            found = False
            subkeys = key.split('.')
            for j in range(len(subkeys)):
                subkey = '.'.join(subkeys[:j + 1])
                if subkey in val:
                    found = True
                    val = val[subkey]
                    keys[0] = '.'.join(subkeys[j + 1:])
                    if not keys[0]:
                        keys.pop(0)
                    break
            if not found:
                val = default
        else:
            val = default

    if expected_type is not None and not isinstance(val, expected_type):
        raise ValueError(
            f'A value of type {expected_type} is expected, "{val}" specified.')

    if raw is True:
        return val

    if isinstance(val, str):
        local_dict.update(custom_dict)
        return cfg_interpolate(val, local_dict)
    elif isinstance(val, dict):

        def interpolate_dict(item, custom_dict):
            res = {}
            for k, v in item.items():
                if allowed_keys and k not in allowed_keys:
                    continue
                if excluded_keys and k in excluded_keys:
                    continue
                if raw_keys and k in raw_keys:
                    res[k] = v
                    continue
                if expand_keys and k not in expand_keys:
                    res[k] = v
                    continue
                if isinstance(v, dict):
                    # v should be processed in place
                    res[k] = interpolate_dict(v, custom_dict)
                elif isinstance(v, str):
                    if '{' in v and '}' in v:
                        local_dict = copy.deepcopy(item)
                        local_dict.update(custom_dict)
                        res[k] = cfg_interpolate(v, local_dict)
                    else:
                        res[k] = v
                else:
                    res[k] = v
            return res

        return interpolate_dict(val, custom_dict)
    else:
        return val


def get_accessed(node):
    '''Get names, but ignore variables names to the left hand side
    That is to say, in case of
       a = b + 1
    we consider b as "being accessed", while a is not.
    '''
    if isinstance(node, ast.Assign):
        return get_accessed(node.value)
    elif isinstance(node, ast.Name):
        return {node.id}
    names = set()
    if isinstance(node, list):
        for x in node:
            names |= get_accessed(x)
    else:
        for x in ast.iter_child_nodes(node):
            names |= get_accessed(x)
    return names


def accessed_vars(statement: str, mode: str = 'exec') -> Set[str]:
    '''Parse a Python statement and analyze the symbols used. The result
    will be used to determine what variables a step depends upon.'''
    try:
        if mode == 'exec':
            return get_accessed(ast.parse(statement, '<string>', 'exec'))
        else:
            res = get_accessed(
                ast.parse('__NULL__(' + statement + ')', '<string>', 'eval'))
            res.remove('__NULL__')
            return res
    except Exception:
        raise RuntimeError(
            f'Failed to parse statement: {statement} in {mode} mode')


def get_used_in_func(node):
    '''Get names, but ignore variables names to the left hand side
    That is to say, in case of
       a = b + 1
    we consider b as "being accessed", while a is not.
    '''
    if isinstance(node, ast.FunctionDef):
        return {node.name: get_accessed(node.body)}
    names = {}
    for node in ast.iter_child_nodes(node):
        names.update(get_used_in_func(node))
    return names


def used_in_func(statement: str,
                 filename: str = '<string>',
                 mode: str = 'exec'):
    '''Parse a Python statement and analyze the symbols used. The result
    will be used to determine what variables a step depends upon.'''
    try:
        return get_used_in_func(ast.parse(statement, filename, mode))
    except Exception as e:
        raise RuntimeError(f'Failed to parse statement: {statement} {e}')


def SoS_eval(expr: str, extra_dict: dict = {}) -> Any:
    '''Evaluate an expression with sos dict.'''
    return eval(expr, env.sos_dict.dict(), extra_dict)


def _is_expr(expr):
    try:
        compile(expr, '<string>', 'eval')
        return True
    except Exception:
        return False


class StatementHash(object):
    stmt_hash = {}

    def __init__(self) -> None:
        pass

    def hash(self, script: str) -> str:
        h = hash(script) & sys.maxsize
        StatementHash.stmt_hash[h] = script
        return f'script_{h}'

    def script(self, hash: str) -> str:
        return StatementHash.stmt_hash[int(hash[7:])]


stmtHash = StatementHash()


def SoS_exec(script: str,
             _dict: dict = None,
             return_result: bool = True) -> None:
    '''Execute a statement.'''
    if _dict is None:
        _dict = env.sos_dict.dict()

    if not return_result:
        if env.verbosity == 0:
            with contextlib.redirect_stdout(None):
                exec(
                    compile(
                        script, filename=stmtHash.hash(script), mode='exec'),
                    _dict)
        else:
            exec(
                compile(script, filename=stmtHash.hash(script), mode='exec'),
                _dict)
        return None

    try:
        stmts = list(ast.iter_child_nodes(ast.parse(script)))
        if not stmts:
            return
        if isinstance(stmts[-1], ast.Expr):
            # the last one is an expression and we will try to return the results
            # so we first execute the previous statements
            if len(stmts) > 1:
                if env.verbosity == 0:
                    with contextlib.redirect_stdout(None):
                        exec(
                            compile(
                                ast.Module(body=stmts[:-1], type_ignores=[]),
                                filename=stmtHash.hash(script),
                                mode="exec"), _dict)
                else:
                    exec(
                        compile(
                            ast.Module(body=stmts[:-1], type_ignores=[]),
                            filename=stmtHash.hash(script),
                            mode="exec"), _dict)
            # then we eval the last one
            if env.verbosity == 0:
                with contextlib.redirect_stdout(None):
                    res = eval(
                        compile(
                            ast.Expression(body=stmts[-1].value),
                            filename=stmtHash.hash(script),
                            mode="eval"), _dict)
            else:
                res = eval(
                    compile(
                        ast.Expression(body=stmts[-1].value),
                        filename=stmtHash.hash(script),
                        mode="eval"), _dict)
        else:
            # otherwise we just execute the entire code
            if env.verbosity == 0:
                with contextlib.redirect_stdout(None):
                    exec(
                        compile(
                            script, filename=stmtHash.hash(script),
                            mode='exec'), _dict)
            else:
                exec(
                    compile(
                        script, filename=stmtHash.hash(script), mode='exec'),
                    _dict)
            res = None
    except SyntaxError as e:
        raise SyntaxError(f"Invalid code {script}: {e}")

    # if check_readonly:
    #    env.sos_dict.check_readonly_vars()
    return res


#
# dynamic expression that cannot be resolved during parsing
# at prepare mode etc, and has to be resolved at run time.
#


class Undetermined(object):

    def __init__(self, expr: str = '') -> None:
        if not isinstance(expr, str):
            raise RuntimeError(
                f'Undetermined expression has to be a string: "{expr}" passed')
        self.expr = expr.strip()

    def value(self):
        return SoS_eval(self.expr)

    def __repr__(self) -> str:
        return f'Undetermined({self.expr!r})'

    def __hash__(self):
        raise RuntimeError(
            'Undetermined expression should be evaluated before used. '
            'This is certainly a bug so please report this to SoS developer.')

    def targets(self) -> 'Undetermined':
        return self


class on_demand_options(object):
    '''Expression that will be evaluated upon request.'''

    def __init__(self, items: Optional[Dict[str, Any]]) -> None:
        self._expressions = {}
        if items:
            self._expressions.update(items)

    def set(self, key: str, value: Any) -> None:
        self._expressions[key] = repr(value)

    def __contains__(self, key: str) -> bool:
        return key in self._expressions

    def __setitem__(self, key: str, value: str) -> None:
        self._expressions[key] = value

    def __getitem__(self, key: str) -> Any:
        # first check if the value if cached
        if key not in self._expressions:
            raise KeyError(key)
        try:
            return SoS_eval(self._expressions[key])
        except Exception as e:
            if key == 'skip':
                raise ValueError(
                    f'Failed to evaluate option {key} with value {self._expressions[key]}: Only constant values are allowed for section option skip'
                )
            else:
                raise ValueError(
                    f'Failed to evaluate option {key} with value {self._expressions[key]}: {e}'
                )

    def __repr__(self):
        return repr(self._expressions)


class KeepOnlyImportAndDefine(ast.NodeTransformer):

    def __init__(self):
        self.level = 0

    def generic_visit(self, node):
        self.level += 1
        if self.level == 2 and not isinstance(
                node,
            (ast.Import, ast.ImportFrom, ast.FunctionDef, ast.ClassDef)):
            # print(f'remove {node}')
            ret = None
        else:
            ret = super(KeepOnlyImportAndDefine, self).generic_visit(node)
        self.level -= 1
        return ret


def analyze_global_statements(global_stmt):
    # find all import and function definition ...
    env.sos_dict.clear()
    env.sos_dict.set('SOS_VERSION', __version__)
    env.sos_dict.set('master_id', '')
    env.sos_dict.set('workflow_id', '')
    env.sos_dict.set('step_name', '')
    env.sos_dict.set('step_id', '')
    # first load CONFIG, this will create CONFIG
    load_config_files()

    # run only import, def, and class of the global_def
    transformer = KeepOnlyImportAndDefine()
    global_def = transformer.visit(
        ast.parse('from sos.runtime import *\n' + global_stmt))
    exec(compile(global_def, filename="<ast>", mode="exec"), env.sos_dict._dict)
    defined_keys = set(env.sos_dict.keys())

    # execute the entire statement
    try:
        SoS_exec(global_stmt)
    except ArgumentError:
        raise
    except Exception as e:
        raise RuntimeError(f'Failed to execute global statement: {e}')
    #
    global_vars = {
        k: env.sos_dict[k] for k in (set(env.sos_dict.keys()) - defined_keys)
        | {'SOS_VERSION', 'CONFIG'}
    }
    # test if global vars can be pickled
    try:
        pickle.dumps(global_vars)
    except Exception:
        for key in set(env.sos_dict.keys()) - defined_keys:
            if not pickleable(env.sos_dict[key], key):
                raise ValueError(
                    f'Variable {key} cannot be defined in global section because it cannot be pickled to workers.'
                )
    return global_def, global_vars



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

#
# Utility functions used by various executors.
#
import os
import sys
import re
import psutil
import traceback

from typing import Any
from collections.abc import Sequence
from io import StringIO
from tokenize import generate_tokens

from .targets import (RemovedTarget, file_target, sos_targets, sos_step, path,
                      dynamic, sos_variable, RuntimeInfo, textMD5)
from .utils import env, Error, format_HHMMSS, expand_size, load_config_files, get_traceback
from .eval import SoS_eval, stmtHash, analyze_global_statements
from .tasks import TaskParams
from .syntax import SOS_TAG
from .controller import request_answer_from_controller


class ExecuteError(Error):
    """An exception to collect exceptions raised during run time so that
    other branches of the DAG would continue if some nodes fail to execute."""

    def __init__(self, workflow: str) -> None:
        Error.__init__(self)
        self.workflow = workflow
        self.errors = []
        self.traces = []
        self.args = (workflow,)

    def append(self, line: str, error: Exception) -> None:
        lines = [x for x in line.split('\n') if x.strip()]
        if not lines:
            short_line = ''
        else:
            short_line = '[' + (lines[0][:40]
                                if len(lines[0]) > 40 else lines[0]) + ']:'
        self.errors.append(short_line)
        self.traces.append(get_traceback())
        newline = '\n' if self.message else ''
        self.message += f'{newline}{short_line} {error}'


def __null_func__(*args, **kwargs) -> Any:
    '''This function will be passed to SoS's namespace and be executed
    to evaluate functions of input, output, and depends directives.'''

    def _flatten(x):
        if isinstance(x, str):
            return [x]
        elif isinstance(x, sos_targets):
            return [x]
        elif isinstance(x, Sequence):
            return sum((_flatten(k) for k in x), [])
        elif hasattr(x, '__flattenable__'):
            return _flatten(x.flatten())
        else:
            return [x]

    return _flatten(args), kwargs


def __output_from__(steps,
                    group_by='auto',
                    paired_with=None,
                    pattern=None,
                    group_with=None,
                    for_each=None,
                    remove_empty_groups=False):
    targets = sos_targets()
    if isinstance(steps, (int, str)):
        steps = [steps]
    elif isinstance(steps, Sequence):
        steps = list(steps)
    else:
        raise ValueError(
            f'Unacceptable value of input prameter from: {steps} provided')
    #
    for step in steps:
        if isinstance(step, int):
            if step == -1:
                # this refers to the last step of a forward style workflow
                if '__last_step__' in env.sos_dict and env.sos_dict[
                        '__last_step__']:
                    step = env.sos_dict['__last_step__']
                else:
                    raise ValueError(
                        f'output_from(-1) is called for a step without previous step'
                    )
            elif '_' in env.sos_dict['step_name']:
                step = f"{env.sos_dict['step_name'].rsplit('_', 1)[0]}_{step}"
            else:
                step = str(step)
        res = request_answer_from_controller(['step_output', step])
        if res is None or not isinstance(res, sos_targets):
            raise RuntimeError(f'Failed to obtain output of step {step}')
        targets.extend(res)

    if group_by != 'auto' or paired_with or pattern or group_with or for_each:
        targets = sos_targets(
            targets,
            group_by=group_by,
            paired_with=paired_with,
            pattern=pattern,
            group_with=group_with,
            for_each=for_each)
    return targets._remove_empty_groups() if remove_empty_groups else targets


def __traced__(*args, **kwargs):
    return sos_targets(*args, **kwargs).set_traced()


def __named_output__(name,
                     group_by='auto',
                     paired_with=None,
                     pattern=None,
                     group_with=None,
                     for_each=None,
                     remove_empty_groups=False):
    targets = request_answer_from_controller(['named_output', name])
    if targets is None:
        env.logger.warning(f'named_output("{name}") is not found')
        return sos_targets([])

    if group_by != 'auto' or paired_with or pattern or group_with or for_each:
        targets = sos_targets(
            targets,
            group_by=group_by,
            paired_with=paired_with,
            pattern=pattern,
            group_with=group_with,
            for_each=for_each)
    return targets._remove_empty_groups() if remove_empty_groups else targets


def clear_output(output=None):
    '''
    Remove file targets in `_output` when a step fails to complete
    '''
    for target in env.sos_dict['_output'] if output is None else output:
        if isinstance(target, file_target) and target.exists():
            try:
                target.unlink()
            except Exception as e:
                env.logger.warning(f'Failed to remove {target}: {e}')


def get_traceback_msg(e):
    error_class = e.__class__.__name__
    tb = sys.exc_info()[-1]
    msg = ''
    for st in reversed(traceback.extract_tb(tb)):
        if st.filename.startswith('script_'):
            code = stmtHash.script(st.filename)
            line_number = st.lineno
            code = '\n'.join([
                f'{"---->" if i+1 == line_number else "     "} {x.rstrip()}'
                for i, x in enumerate(code.splitlines())
            ][max(line_number - 3, 0):line_number + 3])
            msg += f'''\
{st.filename} in {st.name}
{code}
'''
    detail = e.args[0] if e.args else ''
    if msg:
        return f'''
---------------------------------------------------------------------------
{error_class:42}Traceback (most recent call last)
{msg}
{error_class}: {detail}'''
    else:
        return f'{error_class}: {detail}'


def prepare_env(gdef='', gvars={}, extra_vars={}, host='localhost'):
    '''clear current sos_dict, execute global_def (definitions and imports),
    and inject global variables'''
    env.sos_dict.clear()

    if not gdef and not gvars:
        # SoS Notebook calls prepare_env without global statement from a
        # particular
        gdef, gvars = analyze_global_statements('')

    if gdef:
        exec(compile(gdef, filename="<ast>", mode="exec"), env.sos_dict._dict)

    env.sos_dict.quick_update(gvars)
    env.sos_dict.quick_update(extra_vars)
    if 'CONFIG' not in env.sos_dict:
        # if this is in sos notebook
        load_config_files()
    if 'hosts' not in env.sos_dict[
            'CONFIG'] and 'localhost' not in env.sos_dict['CONFIG']:
        env.sos_dict['CONFIG']['localhost'] = 'localhost'
        env.sos_dict['CONFIG']['hosts'] = {
            'localhost': {
                'paths': {},
                'address': 'localhost'
            }
        }
    # expose `paths` of localhost
    if host == 'localhost':
        if 'localhost' in env.sos_dict['CONFIG']:
            if 'hosts' not in env.sos_dict['CONFIG'] or env.sos_dict['CONFIG'][
                    'localhost'] not in env.sos_dict['CONFIG']['hosts']:
                env.logger.warning(
                    f"Localhost {env.sos_dict['CONFIG']['localhost']} is not defined in CONFIG['hosts']"
                )
                env.sos_dict['CONFIG']['hosts'][env.sos_dict['CONFIG']
                                                ['localhost']] = {
                                                    'paths': {},
                                                    'address': 'localhost'
                                                }
            env.sos_dict.set('__host__', env.sos_dict['CONFIG']['localhost'])
        else:
            if 'hosts' in env.sos_dict['CONFIG']:
                if 'localhost' not in env.sos_dict['CONFIG']['hosts']:
                    env.logger.warning('locahost is not defined in "hosts".')
                    env.sos_dict['CONFIG']['hosts']['localhost'] = {
                        'paths': {},
                        'address': 'localhost'
                    }
            elif 'paths' not in env.sos_dict['CONFIG']['hosts']['localhost']:
                env.sos_dict['CONFIG']['hosts']['localhost']['paths'] = {}
            env.sos_dict.set('__host__', 'localhost')
    else:
        if 'hosts' not in env.sos_dict['CONFIG'] or host not in env.sos_dict[
                'CONFIG']['hosts']:
            raise RuntimeError(
                f"Remote host {host} is not defined in CONFIG['hosts']. Available ones are {env.sos_dict['CONFIG']['hosts'].keys()}"
            )
        env.sos_dict.set('__host__', host)


def statementMD5(stmts):

    def _get_tokens(statement):
        return [
            x[1]
            for x in generate_tokens(StringIO(statement).readline)
            if x[1] not in ('', '\n')
        ]

    tokens = []
    for stmt in stmts:
        if stmt:
            tokens.extend(_get_tokens(stmt))
    return textMD5(' '.join(tokens))


def create_task(global_def, global_vars, task_stmt, task_params):
    # env.sos_dict.set('_runtime', {})
    if task_params:
        args, kwargs = SoS_eval(
            f'__null_func__({task_params})',
            extra_dict={'__null_func__': __null_func__})
        if args:
            raise RuntimeError(
                f'Only keyword arguments are accepted for task statement: "{task_params}" provided'
            )
        for k, v in kwargs.items():
            # if k not in SOS_RUNTIME_OPTIONS:
            #     raise RuntimeError(f'Unrecognized runtime option {k}={v}')
            # standardize walltime to an integer
            if k == 'walltime':
                v = format_HHMMSS(v)
            elif k == 'mem':
                v = expand_size(v)
            env.sos_dict['_runtime'][k] = v
    #
    # we need to record the verbosity and sigmode of task during creation because
    # they might be changed while the task is in the queue waiting to be
    # submitted (this happens when tasks are submitted from Jupyter)
    env.sos_dict['_runtime']['verbosity'] = env.verbosity
    env.sos_dict['_runtime']['sig_mode'] = env.config.get('sig_mode', 'default')
    env.sos_dict['_runtime']['run_mode'] = env.config.get('run_mode', 'run')
    if 'workdir' not in env.sos_dict['_runtime']:
        env.sos_dict['_runtime']['workdir'] = path.cwd()
    elif 'TASK' in env.config['SOS_DEBUG'] or 'ALL' in env.config['SOS_DEBUG']:
        env.log_to_file(
            'TASK',
            f'Using specified workdir {env.sos_dict["_runtime"]["workdir"]}')

    # 1324
    env.sos_dict['_runtime'].update(env.config.get('queue_args', {}))

    # NOTE: we do not explicitly include 'step_input', 'step_output',
    # 'step_depends' and 'CONFIG'
    # because they will be included by env.sos_dict['__signature_vars__'] if they are actually
    # used in the task. (issue #752)
    task_vars = env.sos_dict.clone_selected_vars(
        env.sos_dict['__signature_vars__'] | {
            '_input', '_output', '_depends', '_index', 'step_name', '_runtime',
            '__signature_vars__'
        })

    task_tags = [env.sos_dict['step_name'], env.sos_dict['workflow_id'],
        os.path.basename(os.getcwd())]
    if 'tags' in env.sos_dict['_runtime']:
        if isinstance(env.sos_dict['_runtime']['tags'], str):
            tags = [env.sos_dict['_runtime']['tags']]
        elif isinstance(env.sos_dict['_runtime']['tags'], Sequence):
            tags = [str(x) for x in list(env.sos_dict['_runtime']['tags'])]
        else:
            tags = str(env.sos_dict["_runtime"]["tags"]).split(' ')
        #
        for tag in tags:
            if not tag.strip():
                continue
            if not SOS_TAG.match(tag):
                new_tag = re.sub(r'[^\w_.-]', '', tag)
                if new_tag:
                    env.logger.warning(
                        f'Invalid tag "{tag}" is added as "{new_tag}"')
                    task_tags.append(new_tag)
                else:
                    env.logger.warning(f'Invalid tag "{tag}" is ignored')
            else:
                task_tags.append(tag)

    # save task to a file
    taskdef = TaskParams(
        name='{} (index={})'.format(env.sos_dict['step_name'],
                                    env.sos_dict['_index']),
        global_def=(global_def, global_vars),
        task=task_stmt,  # task
        sos_dict=task_vars,
        tags=task_tags)
    # if no output (thus no signature)
    # temporarily create task signature to obtain sig_id
    task_id = RuntimeInfo(
        statementMD5([task_stmt]), task_vars['_input'], task_vars['_output'],
        task_vars['_depends'], task_vars['__signature_vars__'],
        task_vars).sig_id

    # workflow ID should be included but not part of the signature, this is why it is included
    # after task_id is created.
    task_vars['workflow_id'] = env.sos_dict['workflow_id']
    return task_id, taskdef, task_vars


def kill_all_subprocesses(pid=None, include_self=False):
    # kill all subprocesses that could have been spawn from the current process
    try:
        proc = psutil.Process(pid)
    except Exception:
        # if no such process
        return
    procs = proc.children(recursive=True) + ([proc] if include_self else [])
    if not procs:
        return
    for p in procs:
        p.terminate()
    alive = psutil.wait_procs(procs, timeout=3)[-1]
    if alive:
        for p in alive:
            p.kill()
    alive = psutil.wait_procs(procs, timeout=3)[-1]
    if alive:
        for p in alive:
            env.logger.warning(f'Failed to kill subprocess {p.pid}')


def reevaluate_output():
    # re-process the output statement to determine output files
    args, kwargs = SoS_eval(
        f'__null_func__({env.sos_dict["step_output"]._undetermined})',
        extra_dict={
            '__null_func__': __null_func__,
            'output_from': __output_from__,
            'named_output': __named_output__
        })
    if args is True:
        env.logger.error('Failed to resolve unspecified output')
        return
    # handle dynamic args
    args = [x.resolve() if isinstance(x, dynamic) else x for x in args]
    kwargs = {x:(y.resolve() if isinstance(y, dynamic) else y) for x,y in kwargs.items()}
    return sos_targets(*args, **kwargs, _verify_existence=True)


def validate_step_sig(sig):
    if env.config['sig_mode'] in ('default', 'skip', 'distributed'):
        # if users use sos_run, the "scope" of the step goes beyong names in this step
        # so we cannot save signatures for it.
        matched = sig.validate()
        if isinstance(matched, dict):
            env.logger.info(
                f'``{env.sos_dict["step_name"]}`` (index={env.sos_dict["_index"]}) is ``ignored`` due to saved signature'
            )
            return matched
        else:
            env.logger.debug(f'Signature mismatch: {matched}')
            return {}
    elif env.config['sig_mode'] == 'assert':
        matched = sig.validate()
        if isinstance(matched, str):
            raise RuntimeError(f'Signature mismatch: {matched}')
        env.logger.info(
            f'Substep ``{env.sos_dict["step_name"]}`` (index={env.sos_dict["_index"]}) is ``ignored`` with matching signature'
        )
        return matched
    elif env.config['sig_mode'] == 'build':
        # build signature require existence of files
        if sig.write():
            env.logger.info(
                f'Step ``{env.sos_dict["step_name"]}`` (index={env.sos_dict["_index"]}) is ``ignored`` with signature constructed'
            )
            return {
                'input': sig.content['input_obj'],
                'output': sig.content['output_obj'],
                'depends': sig.content['depends_obj'],
                'vars': sig.content['end_context']
            }
    elif env.config['sig_mode'] == 'force':
        return {}
    else:
        raise RuntimeError(
            f'Unrecognized signature mode {env.config["sig_mode"]}')


def strip_param_defs(stmt):
    # the parameters are translated to
    #
    # #begin_parameter name
    # name = sos_get_param("name", value
    # ) #end_parameter name
    #
    # we will need to remove these lines in cases when parameters are not handled
    res = []
    end_line = None
    for line in stmt.splitlines():
        if line.startswith('#begin_parameter '):
            end_line = f') #end_parameter {line[17:]}'
        if not end_line:
            res.append(line)
        elif line == end_line:
            end_line = None
    return '\n'.join(res)


def verify_input(ignore_internal_targets=False):
    # now, if we are actually going to run the script, we
    # need to check the input files actually exists, not just the signatures
    for key in ('_input', '_depends'):
        for target in env.sos_dict[key]:
            if not target.target_exists('target') and not \
                (ignore_internal_targets and isinstance(target, (sos_variable, sos_step))):
                raise RemovedTarget(target)



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import textwrap
from io import StringIO

import pytest

from sos import execute_workflow
from sos.parser import SoS_Script
from sos.targets import file_target
from sos.utils import env
# if the test is imported under sos/test, test interacive executor
from sos.workflow_executor import Base_Executor


def assertDAG(dag, content):
    '''Compare the content of dag in a file, to one of more DAG (strings).'''
    if isinstance(dag, str):
        with open(dag) as d:
            # only get the first DAG
            dot = 'strict' + d.read().split('strict')[1]
    else:
        out = StringIO()
        dag.save(out)
        dot = out.getvalue()

    def sorted_dot(dot):
        return sorted([
            x.strip()
            for x in dot.split('\n')
            if x.strip() and not 'digraph' in x
        ])

    if isinstance(content, str):
        assert sorted_dot(dot) == sorted_dot(content)
    else:
        assert sorted_dot(dot) in [sorted_dot(x) for x in content]


def get_initial_dag(test):
    test = textwrap.dedent(test)
    script = SoS_Script(test)
    wf = script.workflow()
    dag = Base_Executor(wf).initialize_dag()
    return dag


def test_simple_dag(temp_factory):
    '''Test DAG with simple dependency'''
    temp_factory('a.txt', 'a1.txt')
    # basica case
    # 1 -> 2 -> 3 -> 4

    assertDAG(
        get_initial_dag('''
            [A_1]

            [A_2]

            [A_3]

            [A_4]''')
        , textwrap.dedent('''strict digraph "" {
            A_2;
            A_4;
            A_1;
            A_3;
            A_2 -> A_3;
            A_1 -> A_2;
            A_3 -> A_4;
            }
            '''))
    # basica case
    # 1 -> 2 -> 3 -> 4
    assertDAG(
        get_initial_dag('''
            [A_1]

            [A_2]

            [A_3]
            input: 'a.txt'

            [A_4]''')
        , textwrap.dedent('''strict digraph "" {
            A_2;
            A_4;
            A_1;
            A_3;
            A_1 -> A_2;
            A_3 -> A_4;
            }
            '''))
    #
    # 1 -> 2 -> 3 -> 4
    #
    assertDAG(
        get_initial_dag('''
            [A_1]
            input: 'a.txt'
            output: 'b.txt'

            [A_2]
            input: 'b.txt'
            output: 'c.txt'

            [A_3]
            input: 'c.txt'
            output: 'd.txt'

            [A_4]
            input: 'd.txt'
            output: 'e.txt'
            ''')
        , textwrap.dedent('''strict digraph "" {
            A_2;
            A_4;
            A_1;
            A_3;
            A_2 -> A_3;
            A_1 -> A_2;
            A_3 -> A_4;
            }
            '''))
    #
    # 1 -> 2
    # 3 -> 4 (3 does not have any input)
    #
    assertDAG(
        get_initial_dag('''
            [B_1]
            input: 'a.txt'
            output: 'b.txt'

            [B_2]
            input: 'b.txt'
            output: 'c.txt'

            [B_3]
            input: None
            output: 'd.txt'

            [B_4]
            input: 'd.txt'
            output: 'e.txt'
            ''')
        , textwrap.dedent('''strict digraph "" {
            B_2;
            B_4;
            B_1;
            B_3;
            B_1 -> B_2;
            B_3 -> B_4;
            }
            '''))
    #
    # 1 -> 2
    # 3 -> 4 (3 depends on something else)
    #
    assertDAG(
        get_initial_dag('''
            [B_1]
            input: 'a.txt'
            output: 'b.txt'

            [B_2]
            input: 'b.txt'
            output: 'c.txt'

            [B_3]
            input: 'a1.txt'
            output: 'd.txt'

            [B_4]
            input: 'd.txt'
            output: 'e.txt'
            ''')
        , textwrap.dedent('''strict digraph "" {
            B_1;
            B_4;
            B_2;
            B_3;
            B_1 -> B_2;
            B_3 -> B_4;
            }
            '''))
    #
    # (1) -> 2
    # (1) -> 3 -> 4
    #
    # 2 and 3 depends on the output of 1
    assertDAG(
        get_initial_dag('''
            [C_1]
            input: 'a.txt'
            output: 'b.txt'

            [C_2]
            input: 'b.txt'
            output: 'c.txt'

            [C_3]
            input:  'b.txt'
            output: 'd.txt'

            [C_4]
            depends: 'd.txt'
            output: 'e.txt'
            ''')
        , textwrap.dedent('''
            strict digraph "" {
            C_1;
            C_4;
            C_2;
            C_3;
            C_1 -> C_2;
            C_1 -> C_3;
            C_3 -> C_4;
            }
            '''))

def test_undetermined(temp_factory):
    '''Test DAG with undetermined input.'''
    #
    temp_factory('a.txt', 'd.txt')
    # input of step 3 is undertermined so
    # it depends on all its previous steps.
    assertDAG(
        get_initial_dag('''
            [C_1]
            input: 'a.txt'
            output: 'b.txt'

            [C_2]
            input: 'b.txt'
            output: 'c.txt'

            [C_3]
            input:  dynamic('*.txt')
            output: 'd.txt'

            [C_4]
            depends: 'd.txt'
            output: 'e.txt'
            ''')
                # dag.show_nodes()
                # dag.save('a.dot')
        , textwrap.dedent('''
            strict digraph "" {
            C_1;
            C_4;
            C_2;
            C_3;
            C_1 -> C_2;
            C_2 -> C_3;
            C_3 -> C_4;
            }
            '''))
    #
    # output of step
    #
    assertDAG(
        get_initial_dag('''
            [C_1]
            input: 'a.txt'
            output: 'b.txt'

            [C_2]
            input: 'b.txt'
            output: 'c.txt'

            [C_3]
            input:  dynamic('*.txt')

            [C_4]
            depends: 'd.txt'
            output: 'e.txt'
            ''')
        , textwrap.dedent('''
            strict digraph "" {
            C_1;
            C_4;
            C_2;
            C_3;
            C_1 -> C_2;
            C_2 -> C_3;
            C_3 -> C_4;
            }
            '''))

def test_auxiliary_steps(temp_factory,clear_now_and_after):
    graph = textwrap.dedent(('''
        [K: provides='{name}.txt']
        output: f"{name}.txt"

        run: expand=True
        touch '{name}.txt'

        [C_2]
        input: 'b.txt'
        output: 'c.txt'

        run:
        touch c.txt

        [C_3]
        input: 'a.txt'

        '''))
    # a.txt exists and b.txt does not exist
    temp_factory('a.txt')
    clear_now_and_after('b.txt')
    # the workflow should call step K for step C_2, but not C_3
    dag = get_initial_dag(graph)
    #
    # Ticket 363:
    #
    # we have two possibilities here, one is to ignore a.txt,
    # and one is to regenerate a.txt because it is not generated
    # by sos (without signature)        #
    #
    dag.show_nodes()
    assertDAG(
        dag, textwrap.dedent('''
        strict digraph "" {
        "K (b.txt)";
        C_3;
        C_2;
        "K (b.txt)" -> C_2;
        }
        '''))

def test_cycle():
    '''Test cycle detection of DAG'''
    #
    #  A.txt --> B.txt
    #
    #  B.txt --> C.txt
    #
    #  C.txt --> A.txt
    #
    script = SoS_Script(textwrap.dedent('''
    [A_1]
    input: 'A.txt'
    output: 'B.txt'

    [A_2]
    output: 'C.txt'

    [A_3]
    output: 'A.txt'
    '''))
    # the workflow should call step K for step C_2, but not C_3
    wf = script.workflow()
    pytest.raises(RuntimeError, Base_Executor(wf).initialize_dag)

def test_long_chain(clear_now_and_after):
    '''Test long make file style dependencies.'''
    #
    clear_now_and_after('A1.txt', 'A2.txt', 'C2.txt', 'B2.txt', 
        'B1.txt', 'B3.txt', 'C1.txt', 'C3.txt', 'C4.txt')

    #
    #  A1 <- B1 <- B2 <- B3
    #   |
    #   |
    #  \/
    #  A2 <- B2 <- C1 <- C2 <- C4
    #                    C3
    #
    # the workflow should call step K for step C_2, but not C_3
    #env.verbosity = 4
    assertDAG(
        get_initial_dag('''
        [A_1]
        input: 'B1.txt'
        output: 'A1.txt'
        run:
            touch A1.txt

        [A_2]
        depends:  'B2.txt'
        output: 'A2.txt'
        run:
            touch A2.txt

        [B1: provides='B1.txt']
        depends: 'B2.txt'
        run:
            touch B1.txt

        [B2: provides='B2.txt']
        depends: 'B3.txt', 'C1.txt'
        run:
            touch B2.txt

        [B3: provides='B3.txt']
        run:
            touch B3.txt

        [C1: provides='C1.txt']
        depends: 'C2.txt', 'C3.txt'
        run:
            touch C1.txt

        [C2: provides='C2.txt']
        depends: 'C4.txt'
        run:
            touch C2.txt

        [C3: provides='C3.txt']
        depends: 'C4.txt'
        run:
            touch C3.txt

        [C4: provides='C4.txt']
        run:
            touch C4.txt

        ''')
    , textwrap.dedent('''
        strict digraph "" {
        "C4 (C4.txt)";
        "B1 (B1.txt)";
        "C1 (C1.txt)";
        "C2 (C2.txt)";
        "C3 (C3.txt)";
        A_1;
        "B2 (B2.txt)";
        "B3 (B3.txt)";
        A_2;
        "C4 (C4.txt)" -> "C2 (C2.txt)";
        "C4 (C4.txt)" -> "C3 (C3.txt)";
        "B1 (B1.txt)" -> A_1;
        "C1 (C1.txt)" -> "B2 (B2.txt)";
        "C2 (C2.txt)" -> "C1 (C1.txt)";
        "C3 (C3.txt)" -> "C1 (C1.txt)";
        A_1 -> A_2;
        "B2 (B2.txt)" -> "B1 (B1.txt)";
        "B2 (B2.txt)" -> A_2;
        "B3 (B3.txt)" -> "B2 (B2.txt)";
        }
        '''))

def test_target(clear_now_and_after):
    '''Test executing only part of a workflow.'''
    #
    clear_now_and_after('A1.txt', 'A2.txt', 'C2.txt', 'B2.txt', 
        'B1.txt', 'B3.txt', 'C1.txt', 'C3.txt', 'C4.txt')
    #
    #  A1 <- B1 <- B2 <- B3
    #   |
    #   |
    #  \/
    #  A2 <- B2 <- C1 <- C2 <- C4
    #                    C3
    #
    script = SoS_Script(textwrap.dedent('''
    [A1]
    input: 'B1.txt'
    output: 'A1.txt'
    run:
        touch A1.txt

    [A2]
    depends:  'B2.txt'
    run:
        touch A2.txt

    [B1: provides='B1.txt']
    depends: 'B2.txt'
    run:
        touch B1.txt

    [B2: provides='B2.txt']
    depends: 'B3.txt', 'C1.txt'
    run:
        touch B2.txt

    [B3: provides='B3.txt']
    run:
        touch B3.txt

    [C1: provides='C1.txt']
    depends: 'C2.txt', 'C3.txt'
    run:
        touch C1.txt

    [C2: provides='C2.txt']
    depends: 'C4.txt'
    run:
        touch C2.txt

    [C3: provides='C3.txt']
    depends: 'C4.txt'
    run:
        touch C3.txt

    [C4: provides='C4.txt']
    run:
        touch C4.txt

    '''))
    # the workflow should call step K for step C_2, but not C_3
    wf = script.workflow(use_default=False)
    #
    # test 1, we only need to generate target 'B1.txt'
    dag = Base_Executor(wf).initialize_dag(targets=['B1.txt'])
    # note that A2 is no longer mentioned
    assertDAG(
        dag, textwrap.dedent('''
        strict digraph "" {
        "B3 (B3.txt)";
        "C4 (C4.txt)";
        "C2 (C2.txt)";
        "C1 (C1.txt)";
        "B1 (B1.txt)";
        "B2 (B2.txt)";
        "C3 (C3.txt)";
        "B3 (B3.txt)" -> "B2 (B2.txt)";
        "C4 (C4.txt)" -> "C3 (C3.txt)";
        "C4 (C4.txt)" -> "C2 (C2.txt)";
        "C2 (C2.txt)" -> "C1 (C1.txt)";
        "C1 (C1.txt)" -> "B2 (B2.txt)";
        "B2 (B2.txt)" -> "B1 (B1.txt)";
        "C3 (C3.txt)" -> "C1 (C1.txt)";
        }
        '''))
    #
    # test 2, we would like to generate two files
    dag = Base_Executor(wf).initialize_dag(targets=['B2.txt', 'C2.txt'])
    # note that A2 is no longer mentioned
    assertDAG(
        dag, textwrap.dedent('''
        strict digraph "" {
        "C4 (C4.txt)";
        "B2 (B2.txt)";
        "C3 (C3.txt)";
        "B3 (B3.txt)";
        "C2 (C2.txt)";
        "C1 (C1.txt)";
        "C4 (C4.txt)" -> "C2 (C2.txt)";
        "C4 (C4.txt)" -> "C3 (C3.txt)";
        "C3 (C3.txt)" -> "C1 (C1.txt)";
        "B3 (B3.txt)" -> "B2 (B2.txt)";
        "C2 (C2.txt)" -> "C1 (C1.txt)";
        "C1 (C1.txt)" -> "B2 (B2.txt)";
        }
        '''))
    #
    # test 3, generate two separate trees
    #
    dag = Base_Executor(wf).initialize_dag(targets=['B3.txt', 'C2.txt'])
    # note that A2 is no longer mentioned
    assertDAG(
        dag, textwrap.dedent('''
        strict digraph "" {
        "B3 (B3.txt)";
        "C2 (C2.txt)";
        "C4 (C4.txt)";
        "C4 (C4.txt)" -> "C2 (C2.txt)";
        }
        '''))

def test_pattern_reuse(clear_now_and_after):
    '''Test repeated use of steps that use pattern and produce different files.'''
    #
    clear_now_and_after('A1.txt', 'A2.txt', 'B1.txt', 'B1.txt.p', 'B2.txt', 'B2.txt.p')
    #
    #  A1 <- P <- B1
    #  A1 <- P <- B2
    #  A2
    #
    # the workflow should call step K for step C_2, but not C_3
    assertDAG(
        get_initial_dag('''
        [A_1]
        input: 'B1.txt.p', 'B2.txt.p'
        output: 'A1.txt'
        run:
            touch A1.txt

        [A_2]
        output: 'A2.txt'
        run:
            touch A2.txt

        [B1: provides='B1.txt']
        run:
            touch B1.txt

        [B2: provides='B2.txt']
        run:
            touch B2.txt

        [P: provides='{filename}.p']
        input: filename
        run: expand=True
            touch {_output}
        ''')
    , textwrap.dedent('''
        strict digraph "" {
        "P (B2.txt.p)";
        "B1 (B1.txt)";
        "B2 (B2.txt)";
        A_2;
        A_1;
        "P (B1.txt.p)";
        "P (B2.txt.p)" -> A_1;
        "B1 (B1.txt)" -> "P (B1.txt.p)";
        "B2 (B2.txt)" -> "P (B2.txt.p)";
        A_1 -> A_2;
        "P (B1.txt.p)" -> A_1;
        }
        '''))

def test_parallel_execution(clear_now_and_after):
    '''Test basic parallel execution
    A1 <- None
    A2 <- B2
    '''
    clear_now_and_after('A1.txt', 'B2.txt', 'A2.txt')
    # the workflow should call step K for step C_2, but not C_3
    assertDAG(
        get_initial_dag('''
        [A_1]
        output: 'A1.txt'
        run:
            sleep 0
            touch A1.txt

        [A_2]
        input:  'B2.txt'
        output: 'A2.txt'
        run:
            sleep 0
            touch A2.txt

        [B: provides='B2.txt']
        output: 'B2.txt'
        run:
            touch B2.txt

        ''')
    , textwrap.dedent('''
        strict digraph "" {
        A_1;
        A_2;
        "B (B2.txt)";
        "B (B2.txt)" -> A_2;
        }
        '''))
    env.max_jobs = 4
    #env.verbosity = 4
    # the process is slower after switching to spawn mode

def test_shared_dependency(clear_now_and_after):
    #
    # shared variable should introduce additional dependency
    #
    clear_now_and_after('A1.txt')
    #
    # A1 introduces a shared variable ss, A3 depends on ss but not A2
    #
    script = SoS_Script(textwrap.dedent('''
    [A_1: shared='ss']
    ss = 'A1'

    [A_2]
    input: None

    run:
        sleep 0

    [A_3]
    input: None
    import time
    time.sleep(0)
    with open(f"{ss}.txt", 'w') as tmp:
        tmp.write('test')

    '''))
    wf = script.workflow('A')
    dag = Base_Executor(wf).initialize_dag()
    assertDAG(
        dag, textwrap.dedent('''
        strict digraph "" {
        A_3;
        A_1;
        A_2;
        A_1 -> A_3;
        }
        '''))
    env.max_jobs = 3

def test_literal_connection(clear_now_and_after):
    '''Testing the connection of steps with by variables.'''
    clear_now_and_after('A1.txt')
    #
    # A1 introduces a shared variable ss, A3 depends on ss but not A2
    #
    script = SoS_Script(textwrap.dedent('''
    [A_1: shared='p']
    run:
        touch 'A1.txt'

    p = 'A1.txt'

    [A_2]
    input: None

    run:
        sleep 0

    [A_3]
    input: p
    depends: sos_variable('p')

    run:
        sleep 0

    [A_4]
    input: p
    depends: sos_variable('p')
    run:
        sleep 0

    [A_5]
    input: dynamic(p)
    depends: sos_variable('p')
    '''))
    wf = script.workflow('A')
    dag = Base_Executor(wf).initialize_dag()
    assertDAG(
        dag, textwrap.dedent('''
        strict digraph "" {
        A_1;
        A_4;
        A_2;
        A_3;
        A_5;
        A_1 -> A_4;
        A_1 -> A_3;
        A_1 -> A_5;
        }
        '''))
    env.max_jobs = 3

def test_variable_target():
    '''Test dependency caused by variable usage.'''
    script = SoS_Script(textwrap.dedent(r'''
    [A: shared='b']
    b = 1

    [C: shared={'c':'k'}]
    k = 2

    [all: shared='p']
    depends: sos_variable('c'), sos_variable('b')

    p = c + b

    '''))
    wf = script.workflow('all')
    Base_Executor(wf).run()
    assert env.sos_dict['p'] == 3

def test_reverse_shared_variable(clear_now_and_after):
    '''Test shared variables defined in auxiliary steps'''
    clear_now_and_after('a.txt')
    script = SoS_Script(textwrap.dedent(r'''
    [A: shared='b', provides='a.txt']
    b = 1
    run:
    touch a.txt

    [B_1]
    depends: 'a.txt'

    [B_2]
    print(b)

    '''))
    wf = script.workflow('B')
    Base_Executor(wf).run()
    assert env.sos_dict['b'] == 1

def test_chained_depends(temp_factory):
    '''Test chain dependent'''
    temp_factory('a.bam', 'a.bam.bai', 'a.vcf')
    script = SoS_Script(textwrap.dedent(r'''
    # this step provides variable `var`
    [index: provides='{filename}.bam.bai']
    input: f"{filename}.bam"
    run: expand=True
    echo "Generating {_output}"
    touch {_output}

    [call: provides='{filename}.vcf']
    input:   f"{filename}.bam"
    depends: f"{_input}.bai"
    run: expand=True
    echo "Calling variants from {_input} with {_depends} to {_output}"
    touch {_output}
    '''))
    # if file_target('a.bam.bai').exists():
    #     file_target('a.bam.bai').unlink()
    # if file_target('a.vcf').exists():
    #     file_target('a.vcf').unlink()
    # self.touch('a.bam')
    Base_Executor(script.workflow()).run(targets=['a.vcf'])
    # for f in ('a.vcf', 'a.bam', 'a.bam.bai'):
    #     if file_target(f).exists():
    #         file_target(f).unlink()

@pytest.mark.skipif(
    True, reason='This test is failing')
def test_output_of_dag(clear_now_and_after):
    '''Test output of dag'''
    #
    #for f in ['A1.txt', 'A2.txt', 'C2.txt', 'B2.txt', 'B1.txt', 'B3.txt', 'C1.txt', 'C3.txt', 'C4.txt']:
    #    if file_target(f).exists():
    #        file_target(f).unlink()
    #
    #  A1 <- B1 <- B2 <- B3
    #   |
    #   |
    #  \/
    #  A2 <- B2 <- C1 <- C2 <- C4
    #                    C3
    #
    script = SoS_Script(textwrap.dedent('''
    [A_1]
    input: 'B1.txt'
    output: 'A1.txt'
    run:
        touch A1.txt

    [A_2]
    depends:  'B2.txt'
    run:
        touch A2.txt

    [B1: provides='B1.txt']
    depends: 'B2.txt'
    run:
        touch B1.txt

    [B2: provides='B2.txt']
    depends: 'B3.txt', 'C1.txt'
    run:
        touch B2.txt

    [B3: provides='B3.txt']
    run:
        touch B3.txt

    [C1: provides='C1.txt']
    depends: 'C2.txt', 'C3.txt'
    run:
        touch C1.txt

    [C2: provides='C2.txt']
    depends: 'C4.txt'
    run:
        touch C2.txt

    [C3: provides='C3.txt']
    depends: 'C4.txt'
    run:
        touch C3.txt

    [C4: provides='C4.txt']
    run:
        touch C4.txt

    '''))
    # the workflow should call step K for step C_2, but not C_3
    wf = script.workflow(use_default=False)
    #
    # test 1, we only need to generate target 'B1.txt'
    Base_Executor(
        wf,
        config={
            'output_dag': 'test_outofdag1.dot',
            'trace_existing': True
        }).initialize_dag(targets=['B1.txt'])
    # note that A2 is no longer mentioned
    assertDAG(
        'test_outofdag1.dot', textwrap.dedent('''
        strict digraph "" {
        "B3 (B3.txt)";
        "C4 (C4.txt)";
        "C2 (C2.txt)";
        "C1 (C1.txt)";
        "B1 (B1.txt)";
        "B2 (B2.txt)";
        "C3 (C3.txt)";
        "B3 (B3.txt)" -> "B2 (B2.txt)";
        "C4 (C4.txt)" -> "C3 (C3.txt)";
        "C4 (C4.txt)" -> "C2 (C2.txt)";
        "C2 (C2.txt)" -> "C1 (C1.txt)";
        "C1 (C1.txt)" -> "B2 (B2.txt)";
        "B2 (B2.txt)" -> "B1 (B1.txt)";
        "C3 (C3.txt)" -> "C1 (C1.txt)";
        }
        '''))
    # test 2, we would like to generate two files
    Base_Executor(
        wf,
        config={
            'output_dag': 'test_outofdag2.dot',
            'trace_existing': True
        }).initialize_dag(targets=['B2.txt', 'C2.txt'])
    # note that A2 is no longer mentioned
    assertDAG(
        'test_outofdag2.dot', textwrap.dedent('''
        strict digraph "" {
        "C4 (C4.txt)";
        "B2 (B2.txt)";
        "C3 (C3.txt)";
        "B3 (B3.txt)";
        "C2 (C2.txt)";
        "C1 (C1.txt)";
        "C4 (C4.txt)" -> "C2 (C2.txt)";
        "C4 (C4.txt)" -> "C3 (C3.txt)";
        "C3 (C3.txt)" -> "C1 (C1.txt)";
        "B3 (B3.txt)" -> "B2 (B2.txt)";
        "C2 (C2.txt)" -> "C1 (C1.txt)";
        "C1 (C1.txt)" -> "B2 (B2.txt)";
        }
        '''))
    # test 3, generate two separate trees
    #
    Base_Executor(
        wf,
        config={
            'output_dag': 'test_outofdag3.dot',
            'trace_existing': True
        }).initialize_dag(targets=['B3.txt', 'C2.txt'])
    # note that A2 is no longer mentioned
    assertDAG(
        'test_outofdag3.dot', textwrap.dedent('''
        strict digraph "" {
        "B3 (B3.txt)";
        "C2 (C2.txt)";
        "C4 (C4.txt)";
        "C4 (C4.txt)" -> "C2 (C2.txt)";
        }
        '''))
    clear_now_and_after('C2.txt', 'B3.txt', 'C4.txt', 'test.dot', 'test_2.dot')

@pytest.mark.skipif(
    True, reason='This test is failing')
def test_step_with_multiple_output(clear_now_and_after):
    '''Test addition of steps with multiple outputs. It should be added only once'''
    script = SoS_Script(textwrap.dedent('''
    [test_1: provides=['{}.txt'.format(i) for i in range(10)]]
    output: ['{}.txt'.format(i) for i in range(10)]
    run:
    touch {output}

    [test_2: provides=['{}.txt'.format(i) for i in range(10, 20)]]
    depends: ['{}.txt'.format(i) for i in range(10)]
    output: ['{}.txt'.format(i) for i in range(10, 20)]
    run:
    touch {output}

    [default]
    depends: ['{}.txt'.format(i) for i in range(10, 20)]
    '''))
    wf = script.workflow()
    Base_Executor(wf, config={'output_dag': 'test.dot'}).initialize_dag()
    with open('test.dot') as dot:
        lc = len(dot.readlines())
    assert lc == 6
    clear_now_and_after('test.dot')

def test_auxiliary_sos_step():
    '''Testing the use of sos_step with auxiliary step. #736'''
    execute_workflow('''
        [default]
        depends: '1.txt'

        [A_1]
        print("Hi")


        [C_1: provides = "1.txt"]
        depends: sos_step("A_1")
        run:
        touch 1.txt
        ''')

def test_forward_style_depend(clear_now_and_after,temp_factory):
    '''Test the execution of forward-style workflow with undtermined dependency'''
    clear_now_and_after('a.txt.bak')
    temp_factory('a.txt')
    execute_workflow('''
        [10]
        input: 'a.txt'
        output: f"{_input}.bak"
        run: expand=True
            cp {_input} {_output}

        [20]
        depends: "a.txt.bak"
        run: expand=True
            ls {_depends}
        ''')
    assert file_target('a.txt.bak').target_exists()

def test_sos_step_miniworkflow(clear_now_and_after):
    '''Test the addition of mini forward workflows introduced by sos_step'''
    script = SoS_Script(textwrap.dedent('''
    [a_1]
    print(step_name)

    [a_2]
    print(step_name)
    [a_20]
    print(step_name)

    [b_1]
    print(step_name)

    [b_2]
    print(step_name)

    [b_20]
    depends: sos_step('c')
    print(step_name)

    [c_1]
    print(step_name)

    [c_2]
    print(step_name)

    [c_20]
    print(step_name)

    [default]
    depends: sos_step('a'), sos_step('b')
    '''))
    wf = script.workflow()
    Base_Executor(wf, config={'output_dag': 'test.dot'}).run()
    # note that A2 is no longer mentioned
    assertDAG(
        'test.dot', textwrap.dedent('''
        strict digraph "" {
        default;
        a_1;
        a_2;
        a_20;
        b_1;
        b_2;
        b_20;
        c_1;
        c_2;
        c_20;
        a_1 -> a_2;
        a_2 -> a_20;
        a_20 -> default;
        b_1 -> b_2;
        b_2 -> b_20;
        b_20 -> default;
        c_1 -> c_2;
        c_2 -> c_20;
        c_20 -> b_20;
        }
        '''))
    clear_now_and_after('test.dot')

def test_compound_workflow(clear_now_and_after):
    '''Test the DAG of compound workflow'''
    script = SoS_Script(textwrap.dedent('''
    [A_1]
    [A_2]
    [B]
    '''))
    wf = script.workflow('A+B')
    dag = Base_Executor(wf).initialize_dag()
    assertDAG(
        dag, textwrap.dedent('''strict digraph "" {
        A_1;
        A_2;
        B;
        A_1 -> A_2;
        A_2 -> B;
        }'''))
    # with empty depends
    script = SoS_Script(textwrap.dedent('''
    [A_1]
    [A_2]
    [B]
    depends:
    '''))
    wf = script.workflow('A+B')
    dag = Base_Executor(wf).initialize_dag()
    assertDAG(
        dag, textwrap.dedent('''strict digraph "" {
        A_1;
        A_2;
        B;
        A_1 -> A_2;
        A_2 -> B;
        }'''))
    clear_now_and_after('a.txt')
    script = SoS_Script(textwrap.dedent('''
    [A_1]
    [A_2]
    [C]
    output: 'a.txt'
    _output.touch()

    [B]
    depends: 'a.txt'
    '''))
    # with more depends
    wf = script.workflow('A+B')
    dag = Base_Executor(wf).initialize_dag()
    assertDAG(
        dag, textwrap.dedent('''strict digraph "" {
        A_1;
        A_2;
        B;
        "C (a.txt)";
        A_1 -> A_2;
        A_2 -> B;
        "C (a.txt)" -> B;
        }'''))

def test_provides_sos_variable():
    '''Test provides non-filename targets #1341'''
    execute_workflow(
        '''
        [count: provides=sos_variable('numNotebooks')]
        numNotebooks = 1

        [default]
        depends: sos_variable('numNotebooks')
        print(f"There are {numNotebooks} notebooks in this directory")
        ''')


def test_multi_named_output(clear_now_and_after):
    '''Test DAG built from multiple named_output #1166'''
    clear_now_and_after('a.txt', 'b.txt', 'test_named_output.dot')

    execute_workflow(
        '''
        [A]
        output: A='a.txt', B='b.txt'
        _output.touch()

        [default]
        input: named_output('A'), named_output('B')
        ''',
        options={'output_dag': 'test_named_output.dot'})
    # note that A2 is no longer mentioned
    assertDAG('test_named_output.dot', [
        '''
        strict digraph "" {
        default;
        "A (B)";
        "A (B)" -> default;
        }
        ''', '''
        strict digraph "" {
        default;
        "A (A)";
        "A (A)" -> default;
        }
        '''
    ])



#!/usr/bin/env python
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.
import os
import shutil

import pytest

from sos import execute_workflow


def test_bash():
    '''Test action bash'''
    execute_workflow(r'''
        [0]
        bash:
        echo 'Echo'
        ''')


def test_bash_1():
    with pytest.raises(Exception):
        execute_workflow(r'''
            [0]
            bash:
            echo 'Echo
            ''')


def test_sh():
    '''Test action run'''
    execute_workflow(r'''
        [0]
        sh:
        echo 'Echo'
        ''')


def test_sh_1():
    with pytest.raises(Exception):
        execute_workflow(r'''
            [0]
            sh:
            echo 'Echo
            ''')


@pytest.mark.skipif(not shutil.which('csh'), reason="Needs csh")
def test_csh():
    '''Test action csh'''
    execute_workflow(r'''
        [0]
        csh:
            foreach color (red orange yellow green blue)
                echo $color
            end
    ''')


@pytest.mark.skipif(not shutil.which('tcsh'), reason="Needs tcsh")
def test_tcsh():
    '''Test action tcsh'''
    execute_workflow(r'''
        [0]
        tcsh:
            foreach color (red orange yellow green blue)
                echo $color
            end
    ''')


@pytest.mark.skipif(not shutil.which('zsh'), reason="Needs zsh")
def test_zsh():
    '''Test action zsh'''
    execute_workflow(r'''
        [0]
        zsh:
            echo "Hello World!", $SHELL
    ''')


def test_args(clear_now_and_after):
    '''Test args option of scripts'''
    clear_now_and_after('a.txt')

    execute_workflow(r'''
        [0]
        sh: args='-n {filename:q}'
            touch a.txt
    ''')
    assert not os.path.exists('a.txt')
